"""The :mod:`~pyrealm.core.solar` submodule provides functions to calculate
photosynthetic photon flux density (PPFD), daily solar radiation fluxes and other
radiative values.

Several of the calculations share a set of intermediate values (:math:`r_u, r_v, r_w`),
which are calculated using :meth:`calculate_ru_rv_intermediates` and
:meth:`calculate_rw_intermediate`. The functions for these calculations have a public
function that uses standard input variables, and a second private version of the
function that accepts precalculated values of :math:`r_u, r_v, r_w`. These private
functions can be used to avoid repeated calculation of intermediate values where
multiple functions need to be run.
"""  # noqa: D205

from dataclasses import dataclass, field

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import CoreConst
from pyrealm.core.utilities import check_input_shapes


def calculate_distance_factor(
    nu: NDArray[np.float64], solar_eccentricity: float = CoreConst().solar_eccentricity
) -> NDArray[np.float64]:
    r"""Calculates distance factor.

    This function calculates the distance factor :math:`d_r` using the method of
    :cite:t:`berger:1993a`.

    .. math::

        d_r = \left( 1.0 \mathbin{/}
               \left(\frac{1.0 - e^2}
                          {1.0 + e \cos\left(\nu) \right)}
               \right)
             \right)^2

    Args:
        nu: Heliocentric true anomaly (:math:`\nu`, degrees)
        solar_eccentricity: Solar eccentricity (:math:`e`), defaulting to
            :attr:`CoreConst.solar_eccentricity<pyrealm.constants.CoreConst.solar_eccentricity>`.

    Returns:
        An array of distance factors
    """

    return (
        1.0
        / (
            (1.0 - solar_eccentricity**2)
            / (1.0 + solar_eccentricity * np.cos(np.deg2rad(nu)))
        )
    ) ** 2


def calculate_solar_declination_angle(
    lambda_: NDArray[np.float64],
    solar_obliquity: float = CoreConst().solar_obliquity,
) -> NDArray[np.float64]:
    r"""Calculate the solar declination angle.

    This function calculates the solar declination angle (:math:`\delta`, degrees)
    following :cite:t:`Woolf:1968`.

    .. math::

        \delta = \arcsin(\sin(\lambda)\sin(\epsilon))

    Args:
        lambda_: The heliocentric longitude (:math:`\lambda`, degrees)
        solar_obliquity: Solar obliquity (:math:`\epsilon`), defaulting to
            :attr:`CoreConst.solar_obliquity<pyrealm.constants.CoreConst.solar_obliquity>`.

    Returns:
        An array of solar declination angle delta
    """

    return np.rad2deg(
        np.arcsin(np.sin(np.deg2rad(lambda_)) * np.sin(np.deg2rad(solar_obliquity)))
    )


def calculate_ru_rv_intermediates(
    declination: NDArray[np.float64], latitude: NDArray[np.float64]
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    r"""Calculates intermediate values for use in solar radiation calcs.

    This function calculates :math:`r_u` and :math:`r_v` which are dimensionless
    intermediate values calculated from the solar declination angle (:math:`\delta`) and
    the observation latitude (:math:`\phi`).

    .. math::
        :nowrap:

        \[
            \begin{align*}
                r_u &= \sin(\delta) \sin(\phi) \\
                r_v &= \cos(\delta) \cos(\phi)
            \end{align*}
        \]

    Args:
        declination: Solar declination (:math:`\delta`, degrees)
        latitude: Observation latitude (:math:`\phi`, degrees)

    Returns:
        A tuple of :math:`r_u` and :math:`r_v`.
    """
    ru = np.sin(np.deg2rad(declination)) * np.sin(np.deg2rad(latitude))
    rv = np.cos(np.deg2rad(declination)) * np.cos(np.deg2rad(latitude))

    return ru, rv


def calculate_sunset_hour_angle(
    declination: NDArray[np.float64], latitude: NDArray[np.float64]
) -> NDArray[np.float64]:
    r"""Calculate the sunset hour angle.

    This function calculates the sunset hour angle :math:`h_s` in degrees following Eq.
    3.22 of :cite:t:`stine_geyer:2001` as:

    .. math::

        hs = \arccos(-1.0  \frac{r_u}{r_v}),

    where :math:`r_u` and :math:`r_v` are the outputs of
    :meth:`calculate_ru_rv_intermediates`. The function is clamped within :math:`[-180°,
    180°]`.

    Args:
        declination: The solar declination angle (:math:`\delta`, degrees)
        latitude: The site latitude (:math:`\phi`, degrees)

    Returns:
        An array of local hour angle, degrees
    """

    ru, rv = calculate_ru_rv_intermediates(declination=declination, latitude=latitude)

    return _calculate_sunset_hour_angle(ru, rv)


def _calculate_sunset_hour_angle(
    ru: NDArray[np.float64],
    rv: NDArray[np.float64],
) -> NDArray[np.float64]:
    """Calculate sunset hour angle from intermediates.

    This function calculates the sunset hour angle using Eq3.22,
    :cite:t:'stine_geyer:2001', using precalculated intermediate values (see
    :meth:`calculate_ru_rv_intermediates`)

    Args:
        ru: dimensionless parameter
        rv: dimensionless parameter

    Returns:
        An array of local hour angle, degrees
    """

    return np.rad2deg(np.arccos(-1.0 * np.clip(ru / rv, -1.0, 1.0)))


def calculate_daily_solar_radiation(
    distance_factor: NDArray[np.float64],
    sunset_hour_angle: NDArray[np.float64],
    declination: NDArray[np.float64],
    latitude: NDArray[np.float64],
    day_seconds: float = CoreConst().day_seconds,
    solar_constant: float = CoreConst().solar_constant,
) -> NDArray[np.float64]:
    r"""Calculate daily extraterrestrial solar radiation.

    This function calculates the daily extraterrestrial solar radition (:math:`R_d`, J
    m-2) using Eq. 1.10.3 of :cite:t:`Duffie:2013a`:

    .. math::

        R_d = \left( \frac{n_s}{\pi} \right)
            G_{sc} \, d_r \left(r_u \, h_s + r_v \sin(h_s) \right),

    where :math:`r_u` and :math:`r_v` are the outputs of
    :meth:`calculate_ru_rv_intermediates`.



    Args:
        distance_factor: The distance factor (:math:`d_r`, unitless)
        sunset_hour_angle: Local hour angle, (:math:`h_s`, degrees)
        declination: solar declination (:math:`\delta`, degrees)
        latitude: Site latitude (:math:`\phi`, degrees)
        day_seconds: Number of seconds in one solar day (:math:`n_s`, seconds),
            defaulting to
            :attr:`CoreConst.day_seconds<pyrealm.constants.CoreConst.day_seconds>`.
        solar_constant: The solar constant (:math:`G_{sc}`, W m-2), defaulting to
            :attr:`CoreConst.solar_constant<pyrealm.constants.CoreConst.solar_constant>`.

    Returns:
        An array of daily solar radiation, J/m^2
    """

    ru, rv = calculate_ru_rv_intermediates(declination=declination, latitude=latitude)

    return _calculate_daily_solar_radiation(
        ru=ru,
        rv=rv,
        distance_factor=distance_factor,
        sunset_hour_angle=sunset_hour_angle,
        day_seconds=day_seconds,
        solar_constant=solar_constant,
    )


def _calculate_daily_solar_radiation(
    ru: NDArray[np.float64],
    rv: NDArray[np.float64],
    distance_factor: NDArray[np.float64],
    sunset_hour_angle: NDArray[np.float64],
    day_seconds: float = CoreConst().day_seconds,
    solar_constant: float = CoreConst().solar_constant,
) -> NDArray[np.float64]:
    """Calculate daily extraterrestrial solar radiation from intermediate values.

    This function calculates the daily extraterrestrial solar radition (:math:R_d`, J
    m-2) following Eq. 1.10.3 of :cite:t:`Duffie:2013a` and using precalculated
    intermediate values (see :meth:`calculate_ru_rv_intermediates`)

    Args:
        ru: dimensionless variable substitute
        rv: dimensionless variable substitute
        distance_factor: The distance factor (:math:`d_r`, unitless)
        sunset_hour_angle: Local hour angle, (:math:`h_s`, degrees)
        day_seconds: Number of seconds in one solar day (:math:`n_s`, seconds),
            defaulting to
            :attr:`CoreConst.day_seconds<pyrealm.constants.CoreConst.day_seconds>`.
        solar_constant: The solar constant (:math:`G_{sc}`, W m-2), , defaulting to
            :attr:`CoreConst.solar_constant<pyrealm.constants.CoreConst.solar_constant>`.

    Returns:
        An array of daily solar radiation, J/m^2
    """

    return (
        (day_seconds / np.pi)
        * solar_constant
        * distance_factor
        * (
            ru * np.deg2rad(sunset_hour_angle)
            + rv * np.sin(np.deg2rad(sunset_hour_angle))
        )
    )


def calculate_transmissivity(
    sunshine_fraction: NDArray[np.float64],
    elevation: NDArray[np.float64],
    coef: tuple[float, ...] = CoreConst.transmissivity_coef,
) -> NDArray[np.float64]:
    r"""Calculate atmospheric transmissivity.

    This function calculates atmospheric transmissivity (:math:`\tau`, unitless)
    following Eqn. 11 of :cite:t:`Linacre:1968a` and Eq 2 of :cite:t:`allen:1996a`.

    .. math::

        \tau = (c + d \cdot s_f) (1.0 + f \cdot H)

    Args:
        sunshine_fraction: Daily sunshine fraction of observations, (:math:`s_f`,
            unitless)
        elevation: Elevation of observations, (:math:`H`, metres)
        coef: Coefficients of the equation (:math:`c, d, f`), , defaulting to
            :attr:`CoreConst.transmissivity_coef<pyrealm.constants.CoreConst.transmissivity_coef>`.

    Returns:
        An array of bulk transmissivity
    """
    c, d, f = coef
    return (c + d * sunshine_fraction) * (1.0 + f * elevation)


def calculate_ppfd_from_tau_rd(
    transmissivity: NDArray[np.float64],
    daily_solar_radiation: NDArray[np.float64],
    visible_light_albedo: float = CoreConst().visible_light_albedo,
    swdown_to_ppfd_factor: float = CoreConst().swdown_to_ppfd_factor,
) -> NDArray[np.float64]:
    r"""Calculate photosynthetic photon flux density from intermediates.

    This function calculates photosynthetic photon flux density (PPFD, µmol m-2 s-1)
    from precalculated values for daily solar radiation and transimissivity.

    .. math::

        \mathrm{PPFD} = (1.0 \times 10^{-6})  f_{PPFD} (1.0 - A_{vis}) \tau  R_{d}

    Args:
        transmissivity: The bulk transmissivity (:math:`\tau`, unitless)
        daily_solar_radiation: Daily solar radiation (:math:`R_d`, J m-2)
        visible_light_albedo: The visible light albedo (:math:`A_{vis}`, unitless),
            defaulting to
            :attr:`CoreConst.visible_light_albedo<pyrealm.constants.CoreConst.visible_light_albedo>`.
        swdown_to_ppfd_factor: Conversion factor from W m-2 of sunlight to µmol m-2 s-1
            of PPFD (:math:`f_{PPFD}`), defaulting to
            :attr:`CoreConst.swdown_to_ppfd_factor<pyrealm.constants.CoreConst.swdown_to_ppfd_factor>`.

    Returns:
        An array of photosynthetic photon flux density, µmol m-2 s-1.
    """

    ppfd = (
        (1.0e-6)
        * swdown_to_ppfd_factor
        * (1.0 - visible_light_albedo)
        * transmissivity
        * daily_solar_radiation
    )

    return ppfd


def calculate_ppfd(
    sunshine_fraction: NDArray[np.float64],
    elevation: NDArray[np.float64],
    latitude: NDArray[np.float64],
    ordinal_date: NDArray[np.int_],
    n_days: NDArray[np.int_],
    const: CoreConst = CoreConst(),
) -> NDArray[np.float64]:
    r"""Calculates photosynthetic photon flux density (PPFD).

    This function calculates photosynthetic photon flux density (PPFD, mol m-2) for a
    given ordinal date in a year for a location at a given latitude and elevation. The
    PPFD value is moderated by the sunshine fraction for the day to account for effects
    of cloud cover.

    Args:
        sunshine_fraction: Daily sunshine fraction of observations (:math:`s_f`,
            unitless).
        elevation: Elevation of observations, metres.
        latitude: The latitude of observations (degrees).
        ordinal_date: The ordinal_date of the observation.
        n_days: Days in the calendar year.
        const: A CoreConst object.

    Returns:
        An array of photosynthetic photon flux density.

    Example:
        >>> # Define variable values
        >>> # Evaluate function
        >>> calculate_ppfd(
        ...    sunshine_fraction=np.array([1.0]),
        ...    elevation=np.array([142]),
        ...    latitude=np.array([37.7]),
        ...    ordinal_date=np.array([172]),
        ...    n_days=np.array([366])
        ... )
        array([62.04230021])
    """

    # Validate the inputs
    _ = check_input_shapes(latitude, elevation, sunshine_fraction, ordinal_date, n_days)

    # Calculate intermediate values

    nu, lambda_ = calculate_heliocentric_longitudes(
        ordinal_date=ordinal_date,
        n_days=n_days,
        solar_eccentricity=const.solar_eccentricity,
        solar_perihelion=const.solar_perihelion,
    )

    distance_factor = calculate_distance_factor(
        nu=nu, solar_eccentricity=const.solar_eccentricity
    )

    declination = calculate_solar_declination_angle(
        lambda_=lambda_, solar_obliquity=const.solar_obliquity
    )

    tau = calculate_transmissivity(
        sunshine_fraction=sunshine_fraction,
        elevation=elevation,
        coef=const.transmissivity_coef,
    )

    sunset_hour_angle = calculate_sunset_hour_angle(
        declination=declination, latitude=latitude
    )

    ra_d = calculate_daily_solar_radiation(
        distance_factor=distance_factor,
        sunset_hour_angle=sunset_hour_angle,
        declination=declination,
        latitude=latitude,
        day_seconds=const.day_seconds,
        solar_constant=const.solar_constant,
    )

    # Calculate ppfd

    return calculate_ppfd_from_tau_rd(
        transmissivity=tau,
        daily_solar_radiation=ra_d,
        visible_light_albedo=const.visible_light_albedo,
        swdown_to_ppfd_factor=const.swdown_to_ppfd_factor,
    )


def calculate_net_longwave_radiation(
    sunshine_fraction: NDArray[np.float64],
    temperature: NDArray[np.float64],
    coef: tuple[float, float] = CoreConst().net_longwave_radiation_coef,
) -> NDArray[np.float64]:
    r"""Calculate net longwave radiation.

    This function calculates the net longwave radiation (:math:`R_{nl}`, W m-2)
    following  Eqn. 11 and Table 1 of :cite:t:`colinprentice:1993a`:

    .. math::

        R_{nl} = (b + (1.0 - b) n_i) (A - t_c)

    Args:
        sunshine_fraction: Sunshine fraction of observations (:math:`n_i`, unitless)
        temperature: Temperature of observations (:math:`t_c`, °C).
        coef: Coefficients :math:`b` and :math:`A` of the equation,, defaulting to
            :attr:`CoreConst.net_longwave_radiation_coef<pyrealm.constants.CoreConst.net_longwave_radiation_coef>`.

    Returns:
        An array of net longwave radiation.
    """
    b, A = coef
    return (b + (1.0 - b) * sunshine_fraction) * (A - temperature)


def calculate_rw_intermediate(
    transmissivity: NDArray[np.float64],
    distance_factor: NDArray[np.float64],
    shortwave_albedo: float = CoreConst().shortwave_albedo,
    solar_constant: float = CoreConst().solar_constant,
) -> NDArray[np.float64]:
    r"""Calculate the rw intermediate variable.

    This function calculates the widely used variable substitute ``rw`` (:math:`r_w`, W
    m-2) as:

    .. math::

        r_w = (1.0 - A_{sw}) \, \tau  \, G_{sc} \, d_r

    Args:
        transmissivity: Bulk transmissivity, (:math:`\tau`, unitless).
        distance_factor: Distance factor (:math:`d_r`, unitless)
        shortwave_albedo: The shortwave albedo (:math:`A_{sw}`, unitless), defaulting to
            :attr:`CoreConst.shortwave_albedo<pyrealm.constants.CoreConst.shortwave_albedo>`.
        solar_constant: The solar constant (:math:`G_{sc}`, W m-2), defaulting to
            :attr:`CoreConst.solar_constant<pyrealm.constants.CoreConst.solar_constant>`.

    Returns:
        An array of the intermediate variable :math:`r_w` in W m-2.
    """

    return (1.0 - shortwave_albedo) * transmissivity * solar_constant * distance_factor


def calculate_net_radiation_crossover_hour_angle(
    net_longwave_radiation: NDArray[np.float64],
    transmissivity: NDArray[np.float64],
    distance_factor: NDArray[np.float64],
    declination: NDArray[np.float64],
    latitude: NDArray[np.float64],
    shortwave_albedo: float = CoreConst().shortwave_albedo,
    solar_constant: float = CoreConst().solar_constant,
) -> NDArray[np.float64]:
    r"""Calculates the net radiation crossover hour angle.

    This function calculates the net radiation crossover hour angle (:math:`h_n`,
    degrees) as:

    .. math::

        h_n = \arccos(\frac{R_{nl} - r_w  r_u}{r_w  r_v}),

    where :math:`r_u`, :math:`r_v` and :math:`r_w` are the outputs of
    :meth:`calculate_ru_rv_intermediates` and :meth:`calculate_rw_intermediate`. The
    function is clamped within :math:`[-180°, 180°]`.

    Args:
        net_longwave_radiation: Net longwave radiation, (:math:`R_{nl}`, W m-2)
        transmissivity: Bulk transmissivity (:math:`\tau`, unitless)
        declination: solar declination (:math:`\delta`, degrees)
        latitude: Site latitude (:math:`\phi`, degrees)
        distance_factor: The distance factor (:math:`d_r`, unitless)
        shortwave_albedo: The shortwave albedo (:math:`A_{sw}, unitless), defaulting to
            :attr:`CoreConst.shortwave_albedo<pyrealm.constants.CoreConst.shortwave_albedo>`.
        solar_constant: The solar constant, (:math:`G_{sc}`, W m-2), defaulting to
            :attr:`CoreConst.solar_constant<pyrealm.constants.CoreConst.solar_constant>`.

    Returns:
        The net radiation crossover hour angle.
    """

    ru, rv = calculate_ru_rv_intermediates(declination=declination, latitude=latitude)
    rw = calculate_rw_intermediate(
        transmissivity=transmissivity,
        distance_factor=distance_factor,
        shortwave_albedo=shortwave_albedo,
        solar_constant=solar_constant,
    )

    return _calculate_net_radiation_crossover_hour_angle(
        net_longwave_radiation=net_longwave_radiation,
        rw=rw,
        ru=ru,
        rv=rv,
    )


def _calculate_net_radiation_crossover_hour_angle(
    net_longwave_radiation: NDArray[np.float64],
    rw: NDArray[np.float64],
    ru: NDArray[np.float64],
    rv: NDArray[np.float64],
) -> NDArray[np.float64]:
    """Calculate net radiation crossover hour angle using intermediate values.

    This function calculates the net radiation crossover hour angle (:math:`h_n`
    degrees) using precalculated intermediate values (see
    :meth:`calculate_ru_rv_intermediates` and :meth:`calculate_rw_intermediate`)

    Args:
        rw: dimensionless variable substitute
        rv: dimensionless variable substitute
        ru: dimensionless variable substitute
        net_longwave_radiation: net longwave radiation, W/m^2

    Returns:
        An array of net radiation crossover hour angle, degrees
    """

    return np.rad2deg(
        np.arccos(np.clip((net_longwave_radiation - rw * ru) / (rw * rv), -1.0, 1.0))
    )


def calculate_daytime_net_radiation(
    net_longwave_radiation: NDArray[np.float64],
    crossover_hour_angle: NDArray[np.float64],
    declination: NDArray[np.float64],
    latitude: NDArray[np.float64],
    transmissivity: NDArray[np.float64],
    distance_factor: NDArray[np.float64],
    shortwave_albedo: float = CoreConst().shortwave_albedo,
    solar_constant: float = CoreConst().solar_constant,
    day_seconds: float = CoreConst().day_seconds,
) -> NDArray[np.float64]:
    r"""Calculate daily net radiation.

    Calculates the daily net radiation (:math:`R_{nd}`, J m-2) as:

    .. math::

        R_{nd} = \left( \frac{n_s}{\pi} \right)
            \left( h_n  (r_w \, r_u - R_{nl}) + r_w \, r_v \, sin(h_n)) \right)

    where :math:`r_u`, :math:`r_v` and :math:`r_w` are the outputs of
    :meth:`calculate_ru_rv_intermediates` and :meth:`calculate_rw_intermediate`.

    Args:
        net_longwave_radiation: Net longwave radiation, (:math:`R_{nl}`, W m-2)
        crossover_hour_angle: Crossover hour angle, (:math:`h_n`, degrees)
        declination: solar declination (:math:`\delta`, degrees)
        latitude: Site latitude (:math:`\phi`, degrees)
        transmissivity: Bulk transmissivity (:math:`\tau`, unitless)
        distance_factor: The distance factor (:math:`d_r`, unitless)
        shortwave_albedo: The shortwave albedo (:math:`A_{sw}`, unitless), defaulting to
            :attr:`CoreConst.shortwave_albedo<pyrealm.constants.CoreConst.shortwave_albedo>`.
        solar_constant: The solar constant, (:math:`G_{sc}`, W/m^2), defaulting to
            :attr:`CoreConst.solar_constant<pyrealm.constants.CoreConst.solar_constant>`.
        day_seconds: Number of seconds in one solar day (:math:`n_s`, seconds),
            defaulting to
            :attr:`CoreConst.day_seconds<pyrealm.constants.CoreConst.day_seconds>`.

    Result:
        An array of daily net radiation.
    """
    ru, rv = calculate_ru_rv_intermediates(declination=declination, latitude=latitude)
    rw = calculate_rw_intermediate(
        transmissivity=transmissivity,
        distance_factor=distance_factor,
        shortwave_albedo=shortwave_albedo,
        solar_constant=solar_constant,
    )

    return _calculate_daytime_net_radiation(
        rw=rw,
        rv=rv,
        ru=ru,
        crossover_hour_angle=crossover_hour_angle,
        net_longwave_radiation=net_longwave_radiation,
        day_seconds=day_seconds,
    )


def _calculate_daytime_net_radiation(
    rw: NDArray[np.float64],
    rv: NDArray[np.float64],
    ru: NDArray[np.float64],
    crossover_hour_angle: NDArray[np.float64],
    net_longwave_radiation: NDArray[np.float64],
    day_seconds: float = CoreConst().day_seconds,
) -> NDArray[np.float64]:
    """Calculates daily net radiation using precalculated intermediates.

    This function calculates daytime net radiation (:math:`R_{nd}` :math:`J/m^2`), using
    precalculated intermediate values (see :meth:`calculate_ru_rv_intermediates` and
    :meth:`calculate_rw_intermediate`)

    Args:
        rw: variable substitute
        rv: dimensionless variable substitute
        ru: dimensionless variable substitute
        crossover_hour_angle: crossover hour angle (:math:`h_n`, degrees).
        net_longwave_radiation: net longwave radiation (:math:`R_{nl}`, W m-2).
        day_seconds: seconds in one solar day (:math:`n_s`, seconds), defaulting to
            :attr:`CoreConst.day_seconds<pyrealm.constants.CoreConst.day_seconds>`.

    Result:
        An array of daily net radiation, J m-2
    """

    rn_d = (day_seconds / np.pi) * (
        np.deg2rad(crossover_hour_angle) * (rw * ru - net_longwave_radiation)
        + rw * rv * np.sin(np.deg2rad(crossover_hour_angle))
    )

    return rn_d


def calculate_nighttime_net_radiation(
    net_longwave_radiation: NDArray[np.float64],
    crossover_hour_angle: NDArray[np.float64],
    sunset_hour_angle: NDArray[np.float64],
    declination: NDArray[np.float64],
    latitude: NDArray[np.float64],
    transmissivity: NDArray[np.float64],
    distance_factor: NDArray[np.float64],
    shortwave_albedo: float = CoreConst().shortwave_albedo,
    solar_constant: float = CoreConst().solar_constant,
    day_seconds: float = CoreConst().day_seconds,
) -> NDArray[np.float64]:
    r"""Calculates nightime net radiation.

    This function calculates nighttime net radiation (:math:`R_{nn}`, J m-2) as:

    .. math::

        R_{nn} = \left(
            r_w \, r_v \, (\sin(h_s) - \sin(h_n))
            + r_w \, r_u \, (h_s - h_n)
            - R_{nl} \, (\pi -  h_n)
            \right)  \left( \frac{n_s}{\pi} \right),

    where :math:`r_u`, :math:`r_v` and :math:`r_w` are the outputs of
    :meth:`calculate_ru_rv_intermediates` and :meth:`calculate_rw_intermediate`.

    Args:
        net_longwave_radiation: Net longwave radiation, (:math:`R_{nl}`, W m-2)
        crossover_hour_angle: Crossover hour angle, (:math:`h_n`, degrees)
        sunset_hour_angle: sunset hour angle, (:math:`h_s`, degrees)
        declination: solar declination (:math:`\delta`, degrees)
        latitude: Site latitude (:math:`\phi`, degrees)
        transmissivity: Bulk transmissivity (:math:`\tau`, unitless)
        distance_factor: The distance factor (:math:`d_r`, unitless)
        shortwave_albedo: The shortwave albedo (:math:`A_{sw}`, unitless), defaulting to
            :attr:`CoreConst.shortwave_albedo<pyrealm.constants.CoreConst.shortwave_albedo>`.
        solar_constant: The solar constant (:math:`G_{sc}`, W m-2), defaulting to
            :attr:`CoreConst.solar_constant<pyrealm.constants.CoreConst.solar_constant>`.
        day_seconds: Number of seconds in one solar day (:math:`n_s`, seconds),
            defaulting to
            :attr:`CoreConst.day_seconds<pyrealm.constants.CoreConst.day_seconds>`.

    Returns:
        An array of nighttime net radiation in J m-2
    """
    ru, rv = calculate_ru_rv_intermediates(declination=declination, latitude=latitude)
    rw = calculate_rw_intermediate(
        transmissivity=transmissivity,
        distance_factor=distance_factor,
        shortwave_albedo=shortwave_albedo,
        solar_constant=solar_constant,
    )

    return _calculate_nighttime_net_radiation(
        rw=rw,
        rv=rv,
        ru=ru,
        sunset_hour_angle=sunset_hour_angle,
        crossover_hour_angle=crossover_hour_angle,
        net_longwave_radiation=net_longwave_radiation,
        day_seconds=day_seconds,
    )


def _calculate_nighttime_net_radiation(
    rw: NDArray[np.float64],
    rv: NDArray[np.float64],
    ru: NDArray[np.float64],
    sunset_hour_angle: NDArray[np.float64],
    crossover_hour_angle: NDArray[np.float64],
    net_longwave_radiation: NDArray[np.float64],
    day_seconds: float = CoreConst().day_seconds,
) -> NDArray[np.float64]:
    """Calculates nightime net radiation using precalculated intermediates.

    This function calculates nighttime net radiation (:math:`R_{nn}` :math:`J/m^2`), a
    using precalculated intermediate values (see :meth:`calculate_ru_rv_intermediates`
    and :meth:`calculate_rw_intermediate`)

    Args:
        rw: intermediate variable (:math:`r_w`, dimensionless)
        rv: intermediate variable (:math:`r_v`, dimensionless)
        ru: intermediate variable (:math:`r_u`, dimensionless)
        sunset_hour_angle: Sunset hour angle (:math:`h_s`, degrees).
        crossover_hour_angle: Crossover hour angle (:math:`h_n`, degrees)
        net_longwave_radiation: Net longwave radiation, (:math:`r_n`, W m-2)
        day_seconds: Seconds in one solar day (:math:`d_s`, seconds), defaulting to
            :attr:`CoreConst.day_seconds<pyrealm.constants.CoreConst.day_seconds>`.

    Returns:
        An array of nighttime net radiation, :math:`J/m^2`
    """

    return (
        (
            rw
            * rv
            * (
                np.sin(np.deg2rad(sunset_hour_angle))
                - np.sin(np.deg2rad(crossover_hour_angle))
            )
        )
        + (rw * ru * np.deg2rad(sunset_hour_angle - crossover_hour_angle))
        - (net_longwave_radiation * (np.pi - np.deg2rad(crossover_hour_angle)))
    ) * (day_seconds / np.pi)


def calculate_heliocentric_longitudes(
    ordinal_date: NDArray[np.int_],
    n_days: NDArray[np.int_],
    solar_eccentricity: float = CoreConst().solar_eccentricity,
    solar_perihelion: float = CoreConst().solar_perihelion,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    r"""Calculate heliocentric longitude and anomaly.

    This function calculates the heliocentric true anomaly (``nu``, :math:`\nu`,
    degrees) and true longitude (``lambda``, :math:`\lambda`, degrees), given the
    ordinal date in the year and the number of days in the year, following
    :cite:t:`berger:1978a`.

    Args:
        ordinal_date: The ordinal date
        n_days: The number of days in the year
        solar_eccentricity: The solar eccentricity (:math:`e`), defaulting to
            :attr:`CoreConst.solar_eccentricity<pyrealm.constants.CoreConst.solar_eccentricity>`.
        solar_perihelion: The solar perihelion (:math:`\omega`), defaulting to
            :attr:`CoreConst.solar_perihelion<pyrealm.constants.CoreConst.solar_perihelion>`.

    Returns:
        A tuple of NDArrays containing ``nu`` and ``lambda_``.
    """

    # TODO - a lot of wildly unnecessary unit changing here. Can we just work in radians
    #        from the constants up?

    # Variable substitutes:
    eccen_sq = solar_eccentricity**2
    eccen_cb = solar_eccentricity**3
    xse = np.sqrt(1.0 - eccen_sq)

    # Mean longitude for vernal equinox in degrees:
    xlam = np.rad2deg(
        (
            (
                (solar_eccentricity / 2.0 + eccen_cb / 8.0)
                * (1.0 + xse)
                * np.sin(np.deg2rad(solar_perihelion))
            )
            - (
                eccen_sq
                / 4.0
                * (0.5 + xse)
                * np.sin(np.deg2rad(2.0 * solar_perihelion))
            )
            + (
                eccen_cb
                / 8.0
                * (1.0 / 3.0 + xse)
                * np.sin(np.deg2rad(3.0 * solar_perihelion))
            )
        )
        * 2.0
    )

    # Mean longitude for day of year in degrees:
    dlamm = xlam + (ordinal_date - 80.0) * (360.0 / n_days)

    # Mean anomaly in radians:
    ranm = np.deg2rad(dlamm - solar_perihelion)

    # True anomaly in radians:
    ranv = (
        ranm
        + ((2.0 * solar_eccentricity - eccen_cb / 4.0) * np.sin(ranm))
        + (5.0 / 4.0 * eccen_sq * np.sin(2.0 * ranm))
        + (13.0 / 12.0 * eccen_cb * np.sin(3.0 * ranm))
    )

    # True longitude in degrees constrained to 0 - 360
    lambda_ = (np.rad2deg(ranv) + solar_perihelion) % 360

    # True anomaly in degrees constrained to 0 - 360
    nu = (lambda_ - solar_perihelion) % 360

    return nu, lambda_


def calculate_solar_elevation(
    latitude: NDArray[np.float64],
    declination: NDArray[np.float64],
    hour_angle: NDArray[np.float64],
) -> NDArray[np.float64]:
    r"""Calculate the solar elevation angle.

    The solar elevation angle (:math:`\alpha`, radians) is the angle between the horizon
    and the sun, giving the height of the sun in the sky at a given time, following Eqn
    A13 of :cite:t:`depury:1997a`:

    .. math::

        \alpha = \arcsin(\sin(\phi)  \sin(\delta) + \cos(\phi) \cos(\delta) \cos(h))

    Args:
        latitude: Observation latitudes (:math:`\phi`, radians).
        declination: Solar declination angles (:math:`\delta`, radians).
        hour_angle: Observation hour angles (:math:`\h`, radians).

    Returns:
        Solar elevation angles in radians.
    """

    return np.arcsin(
        np.sin(latitude) * np.sin(declination)
        + np.cos(latitude) * np.cos(declination) * np.cos(hour_angle)
    )


def calculate_solar_declination(
    ordinal_date: NDArray[np.int_],
) -> NDArray[np.float64]:
    r"""Calculate solar declination angle.

    Calculates the solar declination angle (:math:`\delta`, radians) from the ordinal
    date, following Eqn A14 of :cite:t:`depury:1997a`.

    .. math::

        \delta = -23.4 \frac{\pi}{180} \cos \left(\frac{2 \pi  (td + 10)}{365}\right)

    Args:
        ordinal_date: The ordinal dates of observations (:math:`t_d`, days).

    Returns:
        Solar declination angles in radians.
    """

    return -23.4 * (np.pi / 180) * np.cos((2 * np.pi * (ordinal_date + 10)) / 365)


def calculate_local_hour_angle(
    current_time: NDArray[np.float64], solar_noon: NDArray[np.float64]
) -> NDArray[np.float64]:
    r"""Calculate the local hour angle.

    The local hour angle (:math:`h`, radians) is a measure of time, expressed as an
    angle indicating the position of the sun relative to solar noon. This function
    calculates the local hour angle following equation A15 of :cite:t:`depury:1997a`.

    .. math::

        h = \pi \frac{t - t_{0}}{12}

    Args:
        current_time: Current time values in decimal hours (:math:`t`).
        solar_noon: Solar noon time values in decimal hours (:math:`t_0`).

    Returns:
        The local hour angle in radians

    """

    return np.pi * (current_time - solar_noon) / 12


def calculate_solar_noon(
    longitude: NDArray[np.float64],
    equation_of_time: NDArray[np.float64],
    standard_longitude: NDArray[np.float64] = np.array([0]),
) -> NDArray[np.float64]:
    r"""Calculate the solar noon  for a given location.

    The solar noon (:math:`t_0`, decimal hour) is the time of day when the sun is at its
    highest point in the sky for a given location. This function calculates the solar
    noon by adjusting the standard noon time (12:00 PM) given the local longitude and
    the equation of time for the day of observation, following Eqn. A16 of
    :cite:t:`depury:1997a`.

    .. math::

        t_{0} = 12 + \frac{4 \cdot -(L_{e} - L_{s}) - E_{t}}{60}

    Args:
        longitude: The local longitude in degrees (:math:`L_e`, degrees).
        equation_of_time: The equation of time given the ordinal date of the
            observation (:math:`E_t`, minutes).
        standard_longitude: The standard meridian for the observation (:math:`L_s`,
            degrees), defaulting to the Greenwich meridan.

    Returns:
        The solar noon time in decimal hours.

    """

    return 12 + (4 * -(longitude - standard_longitude) - equation_of_time) / 60


def calculate_equation_of_time(
    day_angle: NDArray[np.float64],
    coef: tuple[float, ...] = CoreConst.equation_of_time_coef,
) -> NDArray[np.float64]:
    r"""Calculate the equation of time.

    Calculates values of the equation of time (:math:`E_t`, minutes) from the day angle
    as:

    .. math::

        E_t = f \left(
            a + b \cos(\Gamma) + c \sin(\Gamma) + d \cos(2\Gamma) + e \sin(2\Gamma)
            \right),

    where :math:`a,b,c,d,e,f` are the coefficients of the equation provided in the
    ``coef`` argument. The implementation is taken from Eqn 1.4.1 of
    :cite:t:`iqbal:1983a`. Note that Eqn A17 of :cite:t:`depury:1997a` provides an
    implementation that contains errors.

    Args:
        day_angle: The day angle in radians (:math:`\Gamma`, radians)
        coef: A tuple of coefficients for the equation of time, defaulting to the value
            of
            :attr:`CoreConst.equation_of_time_coef<pyrealm.constants.CoreConst.equation_of_time_coef>`.

    Returns:
        An array of equation of time values (minutes)
    """

    a, b, c, d, e, f = coef

    return (
        a
        + b * np.cos(day_angle)
        + c * np.sin(day_angle)
        + d * np.cos(2 * day_angle)
        + e * np.sin(2 * day_angle)
    ) * f


def calculate_day_angle(ordinal_date: NDArray[np.int_]) -> NDArray[np.float64]:
    r"""Calculate the solar day angle.

    Calculates the solar day angle (:math:`\Gamma`, radians) for ordinal dates
    ('Julian dates') using Eqn A18 of :cite:t:`depury:1997a`.

    .. math::

        \Gamma = \frac{2\pi (N - 1)}{365}

    Args:
        ordinal_date: The ordinal date for which to calculate the day angle.

    Returns:
        An array of solar day angles in radians.
    """

    return 2 * np.pi * (ordinal_date - 1) / 365


@dataclass
class SolarPositions:
    """Solar values for observation locations and times.

    This class encapsulates the calculation of key solar position data for observations
    given arrays of the latitude and longitudes of locations and the datetimes of
    observation.

    Example:
        >>> import numpy as np
        >>> sp = SolarPositions(
        ...     latitude=-35.058333,
        ...     longitude=147.34167,
        ...     datetime=np.array([np.datetime64("2024-08-12T10:30:32")]),
        ... )
        >>> sp.decimal_time.round(5)
        array([10.50889])
        >>> sp.solar_elevation.round(5)
        array([0.60252])
    """

    latitude: NDArray[np.float64]
    """The latitude of the location in degrees."""
    longitude: NDArray[np.float64]
    """The longitude of the location in degrees."""
    datetime: NDArray[np.datetime64]
    """An array of np.datetime64 values corresponding to observations at the location
    (local time)."""
    core_const: CoreConst = field(default_factory=lambda: CoreConst())
    """A core constants instance."""

    latitude_rad: NDArray[np.float64] = field(init=False)
    """The latitude of the location in radians, calculated automatically."""
    longitude_rad: NDArray[np.float64] = field(init=False)
    """The longitude of the location in radians, calculated automatically."""
    ordinal_date: NDArray[np.int_] = field(init=False)
    """An array of ordinal dates calculated from the datetimes."""
    decimal_time: NDArray[np.float64] = field(init=False)
    """An array of decimal hour values calculated from the datetimes."""
    local_standard_meridian: NDArray[np.float64] = field(init=False)
    """An array of local meridians given the longitude."""

    day_angle: NDArray[np.float64] = field(init=False)
    r"""The solar day angle of the observations (:math:`\Gamma`, radians)."""
    equation_of_time: NDArray[np.float64] = field(init=False)
    """The equation of time value for the observations (:math:`E_t`, minutes)."""
    solar_noon: NDArray[np.float64] = field(init=False)
    """The solar noon for the observations (:math:`t_0`, decimal hour)."""
    hour_angle: NDArray[np.float64] = field(init=False)
    """The local hour angle for the observations (:math:`h`, radians)."""
    declination: NDArray[np.float64] = field(init=False)
    r"""The declination for the observations ( (:math:`\delta`, radians)."""
    solar_elevation: NDArray[np.float64] = field(init=False)
    r"""The solar elevation for the observations (:math:`\alpha`, radians) ."""

    def __post_init__(self) -> None:
        """Initialise calculated attributes."""

        # Truncate all datetimes to their year to get the first day of each year and
        # then find the difference from the original value to  get the ordinal date
        self.ordinal_date = (
            self.datetime
            - self.datetime.astype("datetime64[Y]").astype("datetime64[D]")
        ).astype("timedelta64[D]").astype("int") + 1

        # Truncate all datetimes to the day to get midnight and then find the difference
        # in seconds and divide through to get decimal hours
        self.decimal_time = (
            self.datetime - self.datetime.astype("datetime64[D]")
        ).astype("timedelta64[s]").astype("int") / 3600

        self.latitude_rad = np.deg2rad(self.latitude)
        self.longitude_rad = np.deg2rad(self.longitude)
        self.local_standard_meridian = self.get_local_standard_meridian()

        self.day_angle = calculate_day_angle(ordinal_date=self.ordinal_date)

        self.equation_of_time = calculate_equation_of_time(
            day_angle=self.day_angle, coef=self.core_const.equation_of_time_coef
        )

        self.solar_noon = calculate_solar_noon(
            longitude=self.longitude,
            equation_of_time=self.equation_of_time,
            standard_longitude=self.local_standard_meridian,
        )

        self.hour_angle = calculate_local_hour_angle(
            current_time=self.decimal_time, solar_noon=self.solar_noon
        )

        self.declination = calculate_solar_declination(ordinal_date=self.ordinal_date)

        self.solar_elevation = calculate_solar_elevation(
            latitude=self.latitude_rad,
            declination=self.declination,
            hour_angle=self.hour_angle,
        )

    def get_local_standard_meridian(self) -> NDArray[np.float64]:
        """Calculate local meridian from longitude.

        This calculates the approximate local standard meridian given the longitude. The
        longitudes are simply mapped to the central longitude of the 15° band in which
        they fall. This is a _very_ approximate estimate of the local time zone.
        """

        return np.floor((self.longitude - -187.5) / 15) * 15 + (-180)



"""The :mod:`~pyrealm.pmodel.competition` submodule provides the
:mod:`~pyrealm.pmodel.competition.C3C4Competition` class, which is used to estimate the
expected fraction of C4 plants given the relative photosynthetic advantages of the two
pathways in locations.
"""  # noqa D210, D415

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import C3C4Const
from pyrealm.core.experimental import warn_experimental
from pyrealm.core.utilities import check_input_shapes, summarize_attrs


def convert_gpp_advantage_to_c4_fraction(
    gpp_adv_c4: NDArray[np.float64],
    treecover: NDArray[np.float64],
    c3c4_const: C3C4Const = C3C4Const(),
) -> NDArray[np.float64]:
    r"""Convert C4 GPP advantage to C4 fraction.

    This function calculates an initial estimate of the fraction of C4 plants based on
    the proportional GPP advantage from C4 photosynthesis. The proportion GPP advantage
    :math:`A_4` is converted to an expected fraction of C4 :math:`F_4` plants using a
    logistic equation of :math:`A_4`, where :math:`A_4` is first modulated by percentage
    tree cover (TC):

    .. math::
        :nowrap:

        \[
            \begin{align*}
                A_4^\prime &= \frac{A_4}{e^ {1 / 1 + \text{TC}}} \\
                F_4 &= \frac{1}{1 + e^{k A_4^\prime} - q}
            \end{align*}
        \]

    The parameters are set in the ``params`` instance and are the slope of the equation
    (:math:`k`, :attr:`~pyrealm.constants.competition_const.C3C4Const.adv_to_frac_k`)
    and :math:`A_4` value at the midpoint of the curve
    (:math:`q`, :attr:`~pyrealm.constants.competition_const.C3C4Const.adv_to_frac_q`).

    Args:
        gpp_adv_c4: The proportional GPP advantage of C4 photosynthesis.
        treecover: The proportion tree cover.
        c3c4_const: Instance of
            :class:`~pyrealm.constants.competition_const.C3C4Const`.

    Returns:
        The estimated fraction of C4 plants given the estimated C4 GPP advantage and
        tree cover.
    """

    frac_c4 = 1.0 / (
        1.0
        + np.exp(
            -c3c4_const.adv_to_frac_k
            * ((gpp_adv_c4 / np.exp(1 / (1 + treecover))) - c3c4_const.adv_to_frac_q)
        )
    )

    return frac_c4


def calculate_tree_proportion(
    gppc3: NDArray[np.float64], c3c4_const: C3C4Const = C3C4Const()
) -> NDArray[np.float64]:
    r"""Calculate the proportion of GPP from C3 trees.

    This function estimates the proportion of C3 trees in the community, which can then
    be used to penalise the fraction of C4 plants due to shading of C4 plants by canopy
    closure, even when C4 photosynthesis is advantagious. The estimated tree cover
    function is:

        .. math::
            :nowrap:

                \[
                    TC(\text{GPP}_{C3}) = a \cdot \text{GPP}_{C3} ^ b - c
                \]

    with parameters set in the `const` instance (:math:`a`,
    :attr:`~pyrealm.constants.competition_const.C3C4Const.gpp_to_tc_a`; :math:`b`,
    :attr:`~pyrealm.constants.competition_const.C3C4Const.gpp_to_tc_b`; :math:`c`,
    :attr:`~pyrealm.constants.competition_const.C3C4Const.gpp_to_tc_c`). The proportion
    of GPP from C3 trees (:math:`h`) is then estimated using the predicted tree cover in
    locations relative to a threshold GPP value (:math:`\text{GPP}_{CLO}`,
    :attr:`~pyrealm.constants.competition_const.C3C4Const.c3_forest_closure_gpp`) above
    which canopy closure occurs. The value of :math:`h` is clamped in :math:`[0, 1]`:

        .. math::
            :nowrap:

                \[
                    h = \max\left(0, \min\left(
                        \frac{TC(\text{GPP}_{C3})}{TC(\text{GPP}_{CLO})}\right),
                        1 \right)
                \]

    Args:
        gppc3: The estimated GPP for C3 plants. The input values here must be
          expressed  as **kilograms** per metre squared per year (kg m-2 yr-1).
        c3c4_const: Instance of
            :class:`~pyrealm.constants.competition_const.C3C4Const`.

    Returns:
        The estimated proportion of GPP resulting from C3 trees.
    """

    prop_trees = (
        c3c4_const.gpp_to_tc_a * np.power(gppc3, c3c4_const.gpp_to_tc_b)
        + c3c4_const.gpp_to_tc_c
    ) / (
        c3c4_const.gpp_to_tc_a
        * np.power(c3c4_const.c3_forest_closure_gpp, c3c4_const.gpp_to_tc_b)
        + c3c4_const.gpp_to_tc_c
    )
    prop_trees = np.clip(prop_trees, 0, 1)

    return prop_trees


class C3C4Competition:
    r"""Implementation of the C3/C4 competition model.

    This class provides an implementation of the calculations of C3/C4 competition,
    described by :cite:t:`lavergne:2020a`. The key inputs ``ggp_c3`` and ``gpp_c4`` are
    gross primary productivity (GPP) estimates for C3 or C4 pathways `alone`  using the
    :class:`~pyrealm.pmodel.pmodel.PModel`

    These estimates are used to calculate the relative advantage of C4 over C3
    photosynthesis (:math:`A_4`), the expected fraction of C4 plants in the community
    (:math:`F_4`) and hence fraction of GPP from C4 plants as follows:

    1. The proportion advantage in GPP for C4 plants is calculated as:

        .. math::
            :nowrap:

            \[
            A_4 = \frac{\text{GPP}_{C4} - \text{GPP}_{C3}}{\text{GPP}_{C3}}
            \]

    2. The proportion GPP advantage :math:`A_4` is converted to an expected fraction of
       C4 :math:`F_4` plants using the function
       :func:`~pyrealm.pmodel.competition.convert_gpp_advantage_to_c4_fraction`.

    3. A model of tree cover from C3 trees is then used to penalise the fraction of C4
       plants due to shading. The function
       :func:`~pyrealm.pmodel.competition.calculate_tree_proportion` is used to estimate
       the proportion (:math:`h`) and the C4 fraction is then discounted as :math:`F_4 =
       F_4 (1 - h)`.

    4. Two masks are applied. First, :math:`F_4 = 0` in locations where the mean  air
       temperature of the coldest month is too low for C4 plants. Second, :math:`F_4` is
       set as unknown for croplands, where the fraction is set by agricultural
       management, not competition.

    Args:
        gpp_c3: Total annual GPP (gC m-2 yr-1) from C3 plants alone.
        gpp_c4: Total annual GPP (gC m-2 yr-1) from C4 plants alone.
        treecover: Percentage tree cover (%).
        below_t_min: A boolean mask, temperatures too low for C4 plants.
        cropland: A boolean mask indicating cropland locations.
        c3c4_const: An instance of
            :class:`~pyrealm.constants.competition_const.C3C4Const` providing
            parameterisation for the competition model.
    """

    # Design Notes: see paper Lavergne et al. (submitted).
    #
    # DO (24/05/2022): I have separated out the functions for different steps
    # into private methods, partly to keep the code cleaner, partly with a
    # slightly hopeful idea that future users could substitute these functions
    # via subclassing, but _mostly_ because being able to access these functions
    # independently makes it much easier to document the steps.

    # TODO - could accept PModel instances for gpp_c3 and gpp_c4 and auto-scale
    #        gpp and check that they are c3 and c4 models.
    #      - Would also allow the estimate isotopic discrimination to work
    #        automatically.
    #      - Axis argument to aggregate values along a time axis?
    #        nansum for gpp  and nanmean for  DeltaC13/4_alone.

    __experimental__ = True

    def __init__(
        self,
        gpp_c3: NDArray[np.float64],
        gpp_c4: NDArray[np.float64],
        treecover: NDArray[np.float64],
        below_t_min: NDArray[np.float64],
        cropland: NDArray[np.float64],
        c3c4_const: C3C4Const = C3C4Const(),
    ):
        warn_experimental("C3C4Competition")

        # Check inputs are congruent
        self.shape: tuple = check_input_shapes(
            gpp_c3, gpp_c4, treecover, cropland, below_t_min
        )
        self.c3c4_const: C3C4Const = c3c4_const

        # Step 1: calculate the percentage advantage in GPP of C4 plants from
        # annual total GPP estimates for C3 and C4 plants. This uses use
        # np.full to handle division by zero without raising warnings
        gpp_adv_c4 = np.full(self.shape, np.nan)
        self.gpp_adv_c4: NDArray[np.float64] = np.divide(
            gpp_c4 - gpp_c3, gpp_c3, out=gpp_adv_c4, where=gpp_c3 > 0
        )
        """The proportional advantage in GPP of C4 over C3 plants"""

        # Step 2: calculate the initial C4 fraction based on advantage modulated
        # by treecover.
        frac_c4 = convert_gpp_advantage_to_c4_fraction(
            self.gpp_adv_c4, treecover=treecover, c3c4_const=c3c4_const
        )

        # Step 3: calculate the proportion of trees shading C4 plants, scaling
        # the predicted GPP to kilograms.
        prop_trees = calculate_tree_proportion(
            gppc3=gpp_c3 / 1000, c3c4_const=c3c4_const
        )
        frac_c4 = frac_c4 * (1 - prop_trees)

        # Step 4: remove areas below minimum temperature
        # mypy - this is a short term fix awaiting better resolution of mixed scalar and
        #        array inputs.
        frac_c4[below_t_min] = 0  # type: ignore

        # Step 5: remove cropland areas
        frac_c4[cropland] = np.nan  # type: ignore

        self.frac_c4: NDArray[np.float64] = frac_c4
        """The estimated fraction of C4 plants."""

        self.gpp_c3_contrib: NDArray[np.float64] = gpp_c3 * (1 - self.frac_c4)
        """The estimated contribution of C3 plants to GPP (gC m-2 yr-1)"""
        self.gpp_c4_contrib = gpp_c4 * self.frac_c4
        """The estimated contribution of C4 plants to GPP (gC m-2 yr-1)"""

        # Define attributes used elsewhere
        self.Delta13C_C3: NDArray[np.float64]
        r"""Contribution from C3 plants to (:math:`\Delta\ce{^13C}`, permil)."""
        self.Delta13C_C4: NDArray[np.float64]
        r"""Contribution from C4 plants to (:math:`\Delta\ce{^13C}`, permil)."""
        self.d13C_C3: NDArray[np.float64]
        r"""Contribution from C3 plants to (:math:`d\ce{^13C}`, permil)."""
        self.d13C_C4: NDArray[np.float64]
        r"""Contribution from C3 plants to (:math:`d\ce{^13C}`, permil)."""

    def __repr__(self) -> str:
        """Generates a string representation of a C3C4Competition instance."""
        return f"C3C4Competition(shape={self.shape})"

    def estimate_isotopic_discrimination(
        self,
        d13CO2: NDArray[np.float64],
        Delta13C_C3_alone: NDArray[np.float64],
        Delta13C_C4_alone: NDArray[np.float64],
    ) -> None:
        r"""Estimate CO2 isotopic discrimination values.

        Creating an instance of :class:`~pyrealm.pmodel.isotopes.CalcCarbonIsotopes`
        from a :class:`~pyrealm.pmodel.pmodel.PModel` instance provides estimated
        total annual descrimination against Carbon 13 (:math:`\Delta\ce{^13C}`) for a
        single photosynthetic pathway.

        This method allows predictions from C3 and C4 pathways to be combined to
        calculate the contribution from C3 and C4 plants given the estimated fraction of
        C4 plants. It also calculates the contributions to annual stable carbon isotopic
        composition (:math:`d\ce{^13C}`).

        Calling this method populates the attributes
        :attr:`~pyrealm.pmodel.competition.C3C4Competition.Delta13C_C3`,
        :attr:`~pyrealm.pmodel.competition.C3C4Competition.Delta13C_C4`,
        :attr:`~pyrealm.pmodel.competition.C3C4Competition.d13C_C3`, and
        :attr:`~pyrealm.pmodel.competition.C3C4Competition.d13C_C4`.

        Args:
            d13CO2: stable carbon isotopic composition of atmospheric CO2
                (permil)
            Delta13C_C3_alone: annual discrimination against 13C for C3
                plants (permil)
            Delta13C_C4_alone: annual discrimination against 13C for C4
                plants (permil)
        """

        _ = check_input_shapes(
            self.gpp_adv_c4, d13CO2, Delta13C_C3_alone, Delta13C_C4_alone
        )

        self.Delta13C_C3 = Delta13C_C3_alone * (1 - self.frac_c4)
        self.Delta13C_C4 = Delta13C_C4_alone * self.frac_c4

        self.d13C_C3 = (d13CO2 - self.Delta13C_C3) / (1 + self.Delta13C_C3 / 1000)
        self.d13C_C4 = (d13CO2 - self.Delta13C_C4) / (1 + self.Delta13C_C4 / 1000)

    def summarize(self, dp: int = 2) -> None:
        """Print summary of estimates of C3/C4 competition.

        Prints a summary of the calculated values in a C3C4Competition instance
        including the mean, range and number of nan values. This will always show
        fraction of C4 and GPP estaimates and isotopic estimates are shown if
        :meth:`~pyrealm.pmodel.competition.C3C4Competition.estimate_isotopic_discrimination`
        has been run.

        Args:
            dp: The number of decimal places used in rounding summary stats.
        """

        attrs: tuple[tuple[str, str], ...] = (
            ("frac_c4", "-"),
            ("gpp_c3_contrib", "gC m-2 yr-1"),
            ("gpp_c4_contrib", "gC m-2 yr-1"),
        )

        if hasattr(self, "d13C_C3"):
            attrs = (
                *attrs,
                ("Delta13C_C3", "permil"),
                ("Delta13C_C4", "permil"),
                ("d13C_C3", "permil"),
                ("d13C_C4", "permil"),
            )

        summarize_attrs(self, attrs, dp=dp)



"""The tmodel_const module TODO."""

from dataclasses import dataclass

from pyrealm.constants import ConstantsClass


@dataclass(frozen=True)
class TModelTraits(ConstantsClass):
    r"""Trait data settings for a TTree instance.

    This data class provides the value of the key traits used in the T model. The
    default values are taken from Table 1 of :cite:t:`Li:2014bc`. Note that the foliage
    maintenance respiration fraction is not named in the T Model description, but has
    been included as a modifiable trait in this implementation. The traits are shown
    below with mathematical notation, default value and units shown in brackets:
    """

    a_hd: float = 116.0
    """Initial slope of height-diameter relationship (:math:`a`, 116.0, -)"""
    ca_ratio: float = 390.43
    """Initial ratio of crown area to stem cross-sectional area
    (:math:`c`, 390.43, -)"""
    h_max: float = 25.33  # H_m, Maximum tree height (m)
    """Maximum tree height (:math:`H_m`, 25.33, m)"""
    rho_s: float = 200.0  # rho_s, Sapwood density (kgCm-3)
    r"""Sapwood density (:math:`\rho_s`, 200.0, kg Cm-3)"""
    lai: float = 1.8  # L, Leaf area index within the crown (-)
    """Leaf area index within the crown (:math:`L`, 1.8, -)"""
    sla: float = 14.0  # sigma, Specific leaf area (m2 kg-1C)
    r"""Specific leaf area (:math:`\sigma`, 14.0, m2 kg-1 C)"""
    tau_f: float = 4.0  # tau_f, Foliage turnover time (years)
    r"""Foliage turnover time (:math:`\tau_f`, 4.0, years)"""
    tau_r: float = 1.04  # tau_r, Fine-root turnover time (years)
    """Fine-root turnover time (:math:`\tau_r`, 1.04, years)"""
    par_ext: float = 0.5  # k, PAR extinction coefficient (-)
    """PAR extinction coefficient (:math:`k`, 0.5, -)"""
    yld: float = 0.6  # y, Yield_factor (-)
    """Yield_factor (:math:`y`, 0.17, -)"""
    zeta: float = 0.17  # zeta, Ratio of fine-root mass to foliage area (kgCm-2)
    r"""Ratio of fine-root mass to foliage area (:math:`\zeta`, 0.17, kg C m-2)"""
    resp_r: float = 0.913  # r_r, Fine-root specific respiration rate (year-1)
    """Fine-root specific respiration rate (:math:`r_r`, 0.913, year-1)"""
    resp_s: float = 0.044  # r_s, Sapwood-specific respiration rate (year-1)
    """Sapwood-specific respiration rate (:math:`r_s`, 0.044, year-1)"""
    resp_f: float = 0.1  # --- , Foliage maintenance respiration fraction (-)
    """Foliage maintenance respiration fraction (:math:`r_f`,  0.1, -)"""

    # TODO: include range + se, or make this another class TraitDistrib
    #       that can yield a Traits instance drawing from that distribution



"""The :mod:`~pyrealm.pmodel.isotopes` submodule provides the
:class:`~pyrealm.pmodel.isotopes.CalcCarbonIsotopes` class, which is used to calculate
isotopic discrimination within the PModel
"""  # noqa D210, D415

from warnings import warn

import numpy as np
from numpy.typing import NDArray

from pyrealm.constants import IsotopesConst
from pyrealm.core.utilities import check_input_shapes, summarize_attrs
from pyrealm.pmodel.pmodel import PModel


class CalcCarbonIsotopes:
    r"""Calculate :math:`\ce{CO2}` isotopic discrimination.

    This class estimates the fractionation of atmospheric CO2 by photosynthetic
    pathways to calculate the isotopic compositions and discrimination given the
    predicted optimal chi from a :class:`~pyrealm.pmodel.pmodel.PModel` instance.

    Discrimination against carbon 13 (:math:`\Delta\ce{^{13}C}`)  is calculated
    using C3 and C4 pathways specific methods, and then discrimination against
    carbon 14 is estimated as :math:`\Delta\ce{^{14}C} \approx 2 \times
    \Delta\ce{^{13}C}` :cite:p:`graven:2020a`. For C3 plants,
    :math:`\Delta\ce{^{13}C}` is calculated both including and excluding
    photorespiration, but these are assumed to be equal for C4 plants. The class
    also reports the isotopic composition of leaves and wood.

    Args:
        pmodel: A :class:`~pyrealm.pmodel.pmodel.PModel` instance providing the
            photosynthetic pathway and estimated optimal chi.
        d13CO2: Atmospheric isotopic ratio for Carbon 13
            (:math:`\delta\ce{^{13}C}`, permil).
        D14CO2: Atmospheric isotopic ratio for Carbon 14
            (:math:`\Delta\ce{^{14}C}`, permil).
        isotopes_const: An instance
            of :class:`~pyrealm.constants.isotope_const.IsotopesConst`, parameterizing
            the calculations.
    """

    def __init__(
        self,
        pmodel: PModel,
        D14CO2: NDArray[np.float64],
        d13CO2: NDArray[np.float64],
        isotopes_const: IsotopesConst = IsotopesConst(),
    ):
        # Check inputs are congruent
        _ = check_input_shapes(pmodel.env.tc, d13CO2, D14CO2)

        self.isotopes_const: IsotopesConst = isotopes_const
        """The IsotopesParams instance used to calculate estimates."""
        self.shape: tuple = pmodel.shape
        """Records the common numpy array shape of array inputs."""
        self.c4: bool = pmodel.c4
        """Indicates if estimates calculated for C3 or C4 photosynthesis."""

        # Attributes defined by methods below
        self.Delta13C_simple: NDArray[np.float64]
        r"""Discrimination against carbon 13 (:math:`\Delta\ce{^{13}C}`, permil)
        excluding photorespiration."""
        self.Delta14C: NDArray[np.float64]
        r"""Discrimination against carbon 13 (:math:`\Delta\ce{^{13}C}`, permil)
        including photorespiration."""
        self.Delta13C: NDArray[np.float64]
        r"""Discrimination against carbon 14 (:math:`\Delta\ce{^{14}C}`, permil)
        including photorespiration."""
        self.d13C_leaf: NDArray[np.float64]
        r"""Isotopic ratio of carbon 13 in leaves
        (:math:`\delta\ce{^{13}C}`, permil)."""
        self.d14C_leaf: NDArray[np.float64]
        r"""Isotopic ratio of carbon 14 in leaves
        (:math:`\delta\ce{^{14}C}`, permil)."""
        self.d13C_wood: NDArray[np.float64]
        r"""Isotopic ratio of carbon 13 in wood (:math:`\delta\ce{^{13}C}`, permil),
        given a parameterized post-photosynthetic fractionation."""

        # Could store pmodel, d13CO2, D14CO2 in instance, but really not needed
        # so try and keep this class simple with a minimum of attributes.
        # TODO: map methods for delta13C to C3 and C4.

        if self.c4:
            self.calc_c4_discrimination(pmodel)
        else:
            self.calc_c3_discrimination(pmodel)

        # 14C discrimination is twice the 13C discrimination (Graven et al. 2020)
        self.Delta14C = self.Delta13C * 2

        # Isotopic composition of leaf
        self.d13C_leaf = (d13CO2 - self.Delta13C) / (1 + self.Delta13C / 1000)
        self.d14C_leaf = (D14CO2 - self.Delta14C) / (1 + self.Delta14C / 1000)

        # Isotopic composition of wood considering post-photosynthetic fractionation:
        self.d13C_wood = self.d13C_leaf + self.isotopes_const.frank_postfrac

    def __repr__(self) -> str:
        """Generates a string representation of a CalcCarbonIsotopes instance."""
        return f"CalcCarbonIsotopes(shape={self.shape}, method={self.c4})"

    def calc_c4_discrimination(self, pmodel: PModel) -> None:
        r"""Calculate C4 isotopic discrimination.

        In this method, :math:`\delta\ce{^{13}C}` is calculated from optimal
        :math:`\chi` using an empirical relationship estimated by
        :cite:p:`lavergne:2022a`.

        Examples:
            >>> import numpy as np
            >>> from pyrealm.pmodel.pmodel import PModel
            >>> from pyrealm.pmodel import PModelEnvironment
            >>> from pyrealm.constants import PModelConst
            >>> pmodel_const = PModelConst(beta_cost_ratio_c4=35)
            >>> env = PModelEnvironment(
            ...     tc=np.array([20]),
            ...     patm=np.array([101325]),
            ...     co2=np.array([400]),
            ...     vpd=np.array([1000]),
            ...     fapar=np.array([1]),
            ...     ppfd=np.array([800]),
            ...     pmodel_const=pmodel_const
            ... )
            >>> mod_c4 = PModel(env, method_optchi='c4_no_gamma')
            >>> mod_c4_delta = CalcCarbonIsotopes(mod_c4, d13CO2= -8.4, D14CO2 = 19.2)
            >>> mod_c4_delta.Delta13C.round(4)
            array([5.6636])
            >>> mod_c4_delta.d13C_leaf.round(4)
            array([-13.9844])
        """

        # Equation from C3/C4 paper
        self.Delta13C_simple = (
            self.isotopes_const.lavergne_delta13_a
            + self.isotopes_const.lavergne_delta13_b * pmodel.optchi.chi
        )
        self.Delta13C = self.Delta13C_simple

    def calc_c4_discrimination_vonC(self, pmodel: PModel) -> None:
        r"""Calculate C4 isotopic discrimination.

        In this method, :math:`\delta\ce{^{13}C}` is calculated from optimal
        :math:`\chi` following Equation 1 in :cite:p:`voncaemmerer:2014a`.

        This method is not yet reachable - it needs a method selection argument to
        switch approaches and check C4 methods are used with C4 pmodels. The method is
        preserving experimental code provided by Alienor Lavergne. A temperature
        sensitive correction term is provided in commented code but not used.

        Examples:
            >>> import numpy as np
            >>> from pyrealm.pmodel.pmodel import PModel
            >>> from pyrealm.pmodel import PModelEnvironment
            >>> from pyrealm.constants import PModelConst
            >>> pmodel_const = PModelConst(beta_cost_ratio_c4=35)
            >>> env = PModelEnvironment(
            ...     tc=np.array([20]),
            ...     patm=np.array([101325]),
            ...     co2=np.array([400]),
            ...     vpd=np.array([1000]),
            ...     fapar=np.array([1]),
            ...     ppfd=np.array([800]),
            ...     pmodel_const=pmodel_const
            ... )
            >>> mod_c4 = PModel(env, method_optchi='c4_no_gamma')
            >>> mod_c4_delta = CalcCarbonIsotopes(mod_c4, d13CO2= -8.4, D14CO2 = 19.2)
            >>> # mod_c4_delta.Delta13C.round(4)
            >>> # array([5.2753])
            >>> # mod_c4_delta.d13C_leaf.round(4)
            >>> # array([-13.6036])
        """

        warn("This method is experimental code from Alienor Lavergne")

        # Equation A5 from von Caemmerer et al. (2014)
        # b4 = (-9.483 * 1000) / (273 + self.tc) + 23.89 + 2.2
        # b4 = self.const.vonCaemmerer_b4

        # 13C discrimination (‰): von Caemmerer et al. (2014) Eq. 1
        self.Delta13C_simple = (
            self.isotopes_const.farquhar_a
            + (
                self.isotopes_const.vonCaemmerer_b4
                + (self.isotopes_const.farquhar_b - self.isotopes_const.vonCaemmerer_s)
                * self.isotopes_const.vonCaemmerer_phi
                - self.isotopes_const.farquhar_a
            )
            * pmodel.optchi.chi
        )

        self.Delta13C = self.Delta13C_simple

    def calc_c3_discrimination(self, pmodel: PModel) -> None:
        r"""Calculate C3 isotopic discrimination.

        This method calculates the isotopic discrimination for
        :math:`\Delta\ce{^{13}C}` both with and without the photorespiratory
        effect following :cite:p:`farquhar:1982a`.

        Examples:
            >>> import numpy as np
            >>> from pyrealm.pmodel import PModelEnvironment
            >>> from pyrealm.pmodel.pmodel import PModel
            >>> env = PModelEnvironment(
            ...              tc=np.array([20]), patm=np.array([101325]),
            ...              co2=np.array([400]), vpd=np.array([1000]),
            ...              fapar=np.array([1]), ppfd=np.array([800]),
            ...              theta=np.array([0.4])
            ... )
            >>> mod_c3 = PModel(env, method_optchi='lavergne20_c3')
            >>> mod_c3_delta = CalcCarbonIsotopes(mod_c3, d13CO2= -8.4, D14CO2 = 19.2)
            >>> mod_c3_delta.Delta13C.round(4)
            array([20.4056])
            >>> mod_c3_delta.d13C_leaf.round(4)
            array([-28.2296])
        """

        # 13C discrimination (permil): Farquhar et al. (1982)
        # Simple
        self.Delta13C_simple = (
            self.isotopes_const.farquhar_a
            + (self.isotopes_const.farquhar_b - self.isotopes_const.farquhar_a)
            * pmodel.optchi.chi
        )

        # with photorespiratory effect:
        self.Delta13C = (
            self.isotopes_const.farquhar_a
            + (self.isotopes_const.farquhar_b2 - self.isotopes_const.farquhar_a)
            * pmodel.optchi.chi
            - self.isotopes_const.farquhar_f * pmodel.env.gammastar / pmodel.env.ca
        )

    def summarize(self, dp: int = 2) -> None:
        """Print summary of values estimated in CalcCarbonIsotopes.

        Prints a summary of the variables calculated within an instance of
        CalcCarbonIsotopes including the mean, range and number of nan values.

        Args:
            dp: The number of decimal places used in rounding summary stats.
        """

        attrs = (
            ("Delta13C_simple", "permil"),  # ‰
            ("Delta13C", "permil"),
            ("Delta14C", "permil"),
            ("d13C_leaf", "permil"),
            ("d14C_leaf", "permil"),
            ("d13C_wood", "permil"),
        )

        summarize_attrs(self, attrs, dp=dp)



"""This module provides shared functionality for the :mod:`~pyrealm.demography` module,
implementing two abstract base classes that are used to share core methods across
demography classes:

* :class:`~pyrealm.demography.core.PandasExporter` provides the utility
  :meth:`~pyrealm.demography.core.PandasExporter.to_pandas` method for extracting data
  from demography classes for plotting and exploring data.
* :class:`~pyrealm.demography.core.CohortMethods` provides the utility
  :meth:`~pyrealm.demography.core.CohortMethods.add_cohort_data` and
  :meth:`~pyrealm.demography.core.CohortMethods.drop_cohort_data` methods that are used
  to append new cohort data across some demography dataclasses.
"""  # noqa: D205, D415

from __future__ import annotations

from abc import ABC
from typing import ClassVar

import numpy as np
import pandas as pd
from numpy.typing import NDArray


class PandasExporter(ABC):
    """Abstract base class implementing pandas export.

    Classes inheriting from this ABC must define a class attribute ``array_attrs`` that
    names a set of class attributes that are all numpy arrays of equal length. The
    classes then inherit the `to_pandas` method that exports those attributes to a
    {class}`pandas.DataFrame`.
    """

    array_attrs: ClassVar[tuple[str, ...]]

    def to_pandas(self) -> pd.DataFrame:
        """Convert the instance array attributes into a {class}`pandas.DataFrame.

        If the array values are two-dimensional (i.e. stem or cohort data by vertical
        heights), the data are stacked and an index is added.
        """

        # Extract the attributes into a dictionary
        data = {k: getattr(self, k) for k in self.array_attrs}

        # Check the first attribute array to see if the values are two dimensional
        data_shape = data[self.array_attrs[0]].shape

        if len(data_shape) == 2:
            # create an index entry to show the column of each value
            stacked_data = {
                "column_stem_index": np.repeat(np.arange(data_shape[1]), data_shape[0])
            }
            # Ravel the attribute data using column-major Fortan style
            for ky, vl in data.items():
                stacked_data[ky] = np.ravel(vl, order="F")

            return pd.DataFrame(stacked_data).set_index("column_stem_index")

        return pd.DataFrame(data)


class CohortMethods(ABC):
    """Abstract base class implementing cohort manipulation functionality.

    Classes inheriting from this ABC must define a class attribute ``array_attrs`` that
    names a set of instance attributes that are all numpy arrays of equal length. They
    must also define the class attribute ``count_attr`` that identifies an instance
    attribute used to records the number of cohorts or stems in the class.

    The class then inherit:

    * The `add_cohorts` method, which allows a second instance of the same class to be
      joined to the calling instance, concatenting each of the array attributes from
      the second instance onto the calling instance and updating ``n_cohorts``.
    * The `drop_cohorts` method, which takes a set of indices onto the array attributes
      and drops the values from those indices for each array attribute and updating
      ``n_cohorts``.
    """

    array_attrs: ClassVar[tuple[str, ...]]
    count_attr: ClassVar[str]

    def add_cohort_data(self, new_data: CohortMethods) -> None:
        """Add array attributes from a second instance implementing the base class.

        Args:
            new_data: A second instance from which to add cohort data to array attribute
                values.
        """

        # Check the incoming dataclass matches the calling instance.
        if not isinstance(new_data, self.__class__):
            raise ValueError(
                f"Cannot add cohort data from an {type(new_data).__name__} "
                f"instance to {type(self).__name__}"
            )

        # Concatenate the array attributes from the incoming instance to the calling
        # instance. This need to respect the attribute array dimensions. If the
        # attribute is one dimensional (e.g. traits), then concatenate on axis=0, but
        # for 2 traits, need to concatenate on axis 1 to extend the trait axis.
        for trait in self.array_attrs:
            current = getattr(self, trait)
            setattr(
                self,
                trait,
                np.concatenate(
                    [current, getattr(new_data, trait)], axis=(current.ndim - 1)
                ),
            )

        # Update the count attribute
        setattr(
            self,
            self.count_attr,
            getattr(self, self.count_attr) + getattr(new_data, self.count_attr),
        )

    def drop_cohort_data(self, drop_indices: NDArray[np.int_]) -> None:
        """Drop array attribute values from an instance.

        Args:
            drop_indices: An array of integer indices to drop from each array attribute.
        """

        # Drop from each trait along the last dimension - handles 2D height x stem and
        # 1D stem traits.

        for trait in self.array_attrs:
            current = getattr(self, trait)
            setattr(
                self, trait, np.delete(current, drop_indices, axis=(current.ndim - 1))
            )

        # Update the count attribute
        setattr(
            self,
            self.count_attr,
            getattr(self, self.count_attr) - len(drop_indices),
        )


def _validate_demography_array_arguments(
    trait_args: dict[str, NDArray],
    size_args: dict[str, NDArray] = {},
    at_size_args: dict[str, NDArray] = {},
) -> None:
    """Shared validation for demography inputs.

    Trait arguments of functions should always be 1 dimensional arrays (row arrays) or
    arrays of size 1 ('scalar'), representing trait values that are constant within a
    cohort or stem. If multiple traits are being validated, then they should all have
    have the same shape or be scalar.

    In addition to simple validation of trait inputs, many demographic functions make
    predictions of stem allometries, allocation and crown profile at a range of sizes.
    These size arguments provide values at which to estimate predictions across stems
    with different traits and  could represent different stem sizes (e.g. dbh) or
    different stem heights at which to evaluate crown profiles. If size arguments are
    provided, then they are also validated. All size arguments must have the same shape
    or be scalar. Given that the stem traits provide N values, the shape of the size
    arguments can then be:

    * A scalar array (i.e. with size 1, such as np.array(1) or np.array([1])), that
      provides a single size at which to calculate predictions for all N stems.
    * A one dimensional array with identical shape to the stem properties (``(N,)``)
      that provides individual single values for each stem.
    * A two dimensional column vector (i.e. with shape ``(M, 1)``) that provides a set
      of ``M`` values at which to calculate multiple predictions for all stem traits.
    * A two-dimensional array ``(M, N)`` that provides individual predictions for each
      stem.

    Lastly, some functions generate predictions for stems at a particular size, given a
    third input. For example, the stem allocation process evaluates how potential GPP is
    allocated for a stem of a given size. (TBD - add crown example). If provided, these
    ``at_size_args`` values are validated as follows:

    * if ``z`` is a row array, ``q_z`` must then have identical shape, or
    * if ``z`` is a column array ``(N, 1)``, ``q_z`` must then have shape ``(N,
        n_stem_properties``).

    Args:
        trait_args: A dictionary of row arrays representing trait values, keyed by the
            trait names.
        size_args: A dictionary of arrays representing size values for stem allometry or
            canopy height at which to evaluate functions, keyed by the value name.
        at_size_args: A dictionary of arrays providing values that are to be evaluated
            given predictions for a set of traits at a given size and which must be
            congruent with both trait and size arguments. The dictionary should be keyed
            with the value name.
    """

    # NOTE - this validation deliberately does not use check_input_shapes because it is
    # insisting on _identical_ shapes or scalar arrays is too restrictive. See
    # discussion in https://github.com/ImperialCollegeLondon/pyrealm/pull/342

    # Check PFT inputs are all equal sized 1D row arrays or a mix of 1D rows and scalar
    # values
    try:
        trait_args_shape = np.broadcast_shapes(
            *[arr.shape for arr in trait_args.values()]
        )
    except ValueError:
        raise ValueError(
            f"Trait arguments are not equal shaped or "
            f"scalar: {','.join(trait_args.keys())}"
        )

    if len(trait_args_shape) > 1:
        raise ValueError(
            f"Trait arguments are not 1D arrays: {','.join(trait_args.keys())}"
        )

    # Check the size inputs are broadcastable if they are provided.
    if size_args:
        try:
            size_args_shape = np.broadcast_shapes(
                *[arr.shape for arr in size_args.values()]
            )
        except ValueError:
            raise ValueError(
                f"Size arguments are not equal shaped or "
                f"scalar: {','.join(size_args.keys())}"
            )

        # Test whether the shape of the size args are compatible with the traits args
        try:
            trait_size_shape = np.broadcast_shapes(trait_args_shape, size_args_shape)
        except ValueError:
            raise ValueError(
                f"The array shapes of the trait {trait_args_shape} and "
                f"size {size_args_shape} arguments are not congruent."
            )

    # Now check at size args if provided
    if at_size_args and not size_args:
        raise ValueError("Only provide `at_size_args` when `size_args` also provided.")

    if at_size_args:
        # Are the at_size values broadcastable?
        try:
            at_size_args_shape = np.broadcast_shapes(
                *[arr.shape for arr in at_size_args.values()]
            )
        except ValueError:
            raise ValueError(
                f"At size arguments are not equal shaped or "
                f"scalar: {','.join(at_size_args.keys())}"
            )

        # Are they congruent with the trait and size values.
        try:
            _ = np.broadcast_shapes(trait_size_shape, at_size_args_shape)
        except ValueError:
            raise ValueError(
                f"The broadcast shapes of the trait and size arguments "
                f"{trait_size_shape} are not congruent with the shape of the at_size "
                f"arguments {at_size_args_shape}."
            )


def _enforce_2D(array: NDArray) -> NDArray:
    """Utility conversion to force two dimensional outputs.

    Depending on the input dimensions, the calculations in the T Model and other
    demography models can return scalar, one dimensional or two dimensional arrays. This
    utility function is used to give a consistent output array dimensionality.

    Args:
        array: A numpy array
    """

    match array.ndim:
        case 0:
            return array[None, None]
        case 1:
            return array[None, :]
        case 2:
            return array
        case _:
            raise ValueError("Demography array of more than 2 dimensions.")



"""The ``t_model`` module provides the basic scaling relationships of the T Model
:cite:`Li:2014bc`. This provides scaling relationships using the plant functional type
traits defined in the :mod:`~pyrealm.demography.flora` module and the diameter at breast
height of individual stems to define the stem geometry, masses, respiration and hence
calculate stem growth given net primary productivity. Note that
:attr:`~pyrealm.demography.tmodel.StemAllometry.stem_height` denotes the total tree
height, as used interchangeable in :cite:`Li:2014bc`, rather than just the height of the
trunk below the canopy.
"""  # noqa: D205

from dataclasses import InitVar, dataclass, field
from typing import ClassVar

import numpy as np
from numpy.typing import NDArray

from pyrealm.core.experimental import warn_experimental
from pyrealm.demography.core import (
    CohortMethods,
    PandasExporter,
    _enforce_2D,
    _validate_demography_array_arguments,
)
from pyrealm.demography.flora import Flora, StemTraits


def calculate_heights(
    h_max: NDArray[np.float64],
    a_hd: NDArray[np.float64],
    dbh: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate tree height under the T Model.

    The height of trees (:math:`H`) are calculated from individual diameters at breast
    height (:math:`D`), along with the maximum height (:math:`H_{m}`) and initial slope
    of the height/diameter relationship (:math:`a`) of the plant functional types
    :cite:p:`{Equation 4, }Li:2014bc`:

    .. math::

         H = H_{m}  \left(1 - \exp(-a \cdot D / H_{m})\right)

    Args:
        h_max: Maximum height of the PFT
        a_hd: Initial slope of the height/diameter relationship of the PFT
        dbh: Diameter at breast height of individuals
        validate: Boolean flag to suppress argument validation
    """

    if validate:
        _validate_demography_array_arguments(
            trait_args={"h_max": h_max, "a_hd": a_hd}, size_args={"dbh": dbh}
        )

    return _enforce_2D(h_max * (1 - np.exp(-a_hd * dbh / h_max)))


def calculate_dbh_from_height(
    h_max: NDArray[np.float64],
    a_hd: NDArray[np.float64],
    stem_height: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate diameter at breast height from stem height under the T Model.

    This function inverts the normal calculation of stem height (:math:`H`) from
    diameter at breast height (DBH, :math:`D`) in the T Model (see
    :meth:`~pyrealm.demography.tmodel.calculate_heights`). This is a helper
    function to allow users to convert known stem heights for a plant functional type,
    with maximum height (:math:`H_{m}`) and initial slope of the height/diameter
    relationship (:math:`a`) into the expected DBH values.

    .. math::

         D = \frac{H \left( \log \left(\frac{H}{H_{m}-H}\right)\right)}{a}

    Warning:
        Where the stem height is greater than the maximum height for a PFT, then
        DBH is undefined and the return array will contain `np.nan`. Where the
        stem height equals the maximum height, the model predicts an infinite stem
        diameter: the `h_max` parameter is the asymptotic maximum stem height of an
        exponential function. Similarly, heights very close to the maximum height may
        lead to unrealistically large predictions of DBH.

    Args:
        h_max: Maximum height of the PFT
        a_hd: Initial slope of the height/diameter relationship of the PFT
        stem_height: Stem height of individuals
        validate: Boolean flag to suppress argument validation
    """

    if validate:
        _validate_demography_array_arguments(
            trait_args={"h_max": h_max, "a_hd": a_hd},
            size_args={"stem_height": stem_height},
        )

    # The equation here blows up in a couple of ways:
    # - H > h_max leads to negative logs which generates np.nan with an invalid value
    #   warning. The np.nan here is what we want to happen, so the warning needs
    #   suppressing.
    # - H = h_max generates a divide by zero which returns inf with a warning. Here the
    #   answer should be h_max so that needs trapping.

    with np.testing.suppress_warnings() as sup:
        sup.filter(RuntimeWarning, "divide by zero encountered in divide")
        sup.filter(RuntimeWarning, "invalid value encountered in log")
        return _enforce_2D((h_max * np.log(h_max / (h_max - stem_height))) / a_hd)


def calculate_crown_areas(
    ca_ratio: NDArray[np.float64],
    a_hd: NDArray[np.float64],
    dbh: NDArray[np.float64],
    stem_height: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate tree crown area under the T Model.

    The tree crown area (:math:`A_{c}`) is calculated from individual diameters at
    breast height (:math:`D`) and stem height (:math:`H`), along with the crown area
    ratio (:math:`c`) and the initial slope of the height/diameter relationship
    (:math:`a`) of the plant functional type :cite:p:`{Equation 8, }Li:2014bc`:

    .. math::

        A_{c} =\frac{\pi c}{4 a} D H


    Args:
        ca_ratio: Crown area ratio of the PFT
        a_hd: Initial slope of the height/diameter relationship of the PFT
        dbh: Diameter at breast height of individuals
        stem_height: Stem height of individuals
        validate: Boolean flag to suppress argument validation
    """

    if validate:
        _validate_demography_array_arguments(
            trait_args={"ca_ratio": ca_ratio, "a_hd": a_hd},
            size_args={"dbh": dbh, "stem_height": stem_height},
        )

    return _enforce_2D(((np.pi * ca_ratio) / (4 * a_hd)) * dbh * stem_height)


def calculate_crown_fractions(
    a_hd: NDArray[np.float64],
    stem_height: NDArray[np.float64],
    dbh: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate tree crown fraction under the T Model.

    The crown fraction (:math:`f_{c}`) is calculated from individual diameters at breast
    height (:math:`D`) and stem height (:math:`H`), along with the initial slope of the
    height / diameter relationship (:math:`a`) of the plant functional type
    :cite:p:`{Equation 11, }Li:2014bc`:

    .. math::

        f_{c} =\frac{H}{a D}

    Args:
        a_hd: Initial slope of the height/diameter relationship of the PFT
        stem_height: Stem height of individuals
        dbh: Diameter at breast height of individuals
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"a_hd": a_hd},
            size_args={"dbh": dbh, "stem_height": stem_height},
        )

    return _enforce_2D(stem_height / (a_hd * dbh))


def calculate_stem_masses(
    rho_s: NDArray[np.float64],
    stem_height: NDArray[np.float64],
    dbh: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate stem mass under the T Model.

    The stem mass (:math:`W_{s}`) is calculated from individual diameters at breast
    height (:math:`D`) and stem height (:math:`H`), along with the wood density
    (:math:`\rho_s`) of the plant functional type :cite:p:`{Equation 6, }Li:2014bc`:

    .. math::

        W_s = (\pi / 8) \rho_s D^2 H

    Args:
        rho_s: Wood density of the PFT
        stem_height: Stem height of individuals
        dbh: Diameter at breast height of individuals
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"rho_s": rho_s},
            size_args={"dbh": dbh, "stem_height": stem_height},
        )

    return _enforce_2D((np.pi / 8) * rho_s * (dbh**2) * stem_height)


def calculate_foliage_masses(
    sla: NDArray[np.float64],
    lai: NDArray[np.float64],
    crown_area: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate foliage mass under the T Model.

    The foliage mass (:math:`W_{f}`) is calculated from the crown area (:math:`A_{c}`),
    along with the specific leaf area (:math:`\sigma`) and leaf area index (:math:`L`)
    of the plant functional type :cite:p:`Li:2014bc`.

    .. math::

        W_f = (1 / \sigma) A_c L

    Args:
        sla: Specific leaf area of the PFT
        lai: Leaf area index of the PFT
        crown_area: Crown area of individuals
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"sla": sla, "lai": lai},
            size_args={"crown_area": crown_area},
        )

    return _enforce_2D(crown_area * lai * (1 / sla))


def calculate_sapwood_masses(
    rho_s: NDArray[np.float64],
    ca_ratio: NDArray[np.float64],
    stem_height: NDArray[np.float64],
    crown_area: NDArray[np.float64],
    crown_fraction: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate sapwood mass under the T Model.

    The sapwood mass (:math:`W_{\cdot s}`) is calculated from the individual crown area
    (:math:`A_{c}`), stem height (:math:`H`) and canopy fraction (:math:`f_{c}`) along
    with the wood density (:math:`\rho_s`) and crown area ratio (:math:`c`) of the
    plant functional type :cite:p:`{Equation 14, }Li:2014bc`.

    .. math::

        W_{\cdot s} = \frac{A_c \rho_s H (1 - f_c / 2)}{c}

    Args:
        rho_s: Wood density of the PFT
        ca_ratio: Crown area ratio of the PFT
        stem_height: Stem height of individuals
        crown_area: Crown area of individuals
        crown_fraction: Crown fraction of individuals
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"rho_s": rho_s, "ca_ratio": ca_ratio},
            size_args={
                "stem_height": stem_height,
                "crown_area": crown_area,
                "crown_fraction": crown_fraction,
            },
        )

    return _enforce_2D(
        crown_area * rho_s * stem_height * (1 - crown_fraction / 2) / ca_ratio
    )


def calculate_crown_z_max(
    z_max_prop: NDArray[np.float64],
    stem_height: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate height of maximum crown radius.

    The height of the maximum crown radius (:math:`z_m`) is derived from the crown
    shape parameters (:math:`m,n`) and the resulting fixed proportion (:math:`p_{zm}`)
    for plant functional types. These shape parameters are defined as part of the
    extension of the T Model presented by :cite:t:`joshi:2022a`.

    The value :math:`z_m` is the height above ground where the largest crown radius is
    found, given the proportion and the estimated stem height (:math:`H`) of
    individuals.

    .. math::

        z_m = p_{zm} H

    Args:
        z_max_prop: Crown shape parameter of the PFT
        stem_height: Stem height of individuals
        validate: Boolean flag to suppress argument validation
    """

    if validate:
        _validate_demography_array_arguments(
            trait_args={"z_max_prop": z_max_prop},
            size_args={"stem_height": stem_height},
        )

    return _enforce_2D(stem_height * z_max_prop)


def calculate_crown_r0(
    q_m: NDArray[np.float64],
    crown_area: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate scaling factor for width of maximum crown radius.

    This scaling factor (:math:`r_0`) is derived from the crown shape parameters
    (:math:`m,n,q_m`) for plant functional types and the estimated crown area
    (:math:`A_c`) of individuals. The shape parameters are defined as part of the
    extension of the T Model presented by :cite:t:`joshi:2022a` and :math:`r_0` is used
    to scale the crown area such that the crown area at the  maximum crown radius fits
    the expectations of the T Model.

    .. math::

        r_0 = 1/q_m  \sqrt{A_c / \pi}

    Args:
        q_m: Crown shape parameter of the PFT
        crown_area: Crown area of individuals
        validate: Boolean flag to suppress argument validation

    """

    if validate:
        _validate_demography_array_arguments(
            trait_args={"q_m": q_m},
            size_args={"crown_area": crown_area},
        )

    # Scaling factor to give expected A_c (crown area) at
    # z_m (height of maximum crown radius)
    return _enforce_2D(1 / q_m * np.sqrt(crown_area / np.pi))


def calculate_whole_crown_gpp(
    potential_gpp: NDArray[np.float64],
    crown_area: NDArray[np.float64],
    par_ext: NDArray[np.float64],
    lai: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate whole crown gross primary productivity.

    This function calculates individual gross primary productivity (GPP) across the
    whole crown, given the individual potential GPP per metre squared (:math:`P_0`, kg C
    m-2) and crown area (:math:`A_c`, m2), along with the leaf area index (:math:`L`)
    and the extinction coefficient (:math:`k`) of the plant functional type
    :cite:p:`{Equation 12, }Li:2014bc`.

    .. math::

        P = P_0 A_c (1 - e^{-kL})

    Args:
        lai: The leaf area index
        par_ext: The extinction coefficient
        potential_gpp: Potential GPP per metre squared
        crown_area: The crown area in metres squared
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"lai": lai, "par_ext": par_ext},
            size_args={"potential_gpp": potential_gpp, "crown_area": crown_area},
        )

    return _enforce_2D(potential_gpp * crown_area * (1 - np.exp(-(par_ext * lai))))


def calculate_sapwood_respiration(
    resp_s: NDArray[np.float64],
    sapwood_mass: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate sapwood respiration.

    Calculates the total sapwood respiration (:math:`R_{\cdot s}`) given the individual
    sapwood mass (:math:`W_{\cdot s}`) and the sapwood respiration rate of the plant
    functional type (:math:`r_{s}`) :cite:p:`{see Equation 13, }Li:2014bc`.

    .. math::
         R_{\cdot s} = W_{\cdot s} \, r_s

    Args:
        resp_s: The sapwood respiration rate
        sapwood_mass: The individual sapwood mass
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"resp_s": resp_s},
            size_args={"sapwood_mass": sapwood_mass},
        )

    return _enforce_2D(sapwood_mass * resp_s)


def calculate_foliar_respiration(
    resp_f: NDArray[np.float64],
    whole_crown_gpp: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate foliar respiration.

    Calculates the total foliar respiration (:math:`R_{f}`) given the individual crown
    GPP (:math:`P`) and the foliar respiration rate of the plant functional type
    (:math:`r_{f}`). :cite:t:`Li:2014bc` remove foliar respiration as a constant
    proportion of potential GPP before calculating GPP for the crown, but ``pyrealm``
    treats this proportion as part of the definition of plant functional types.

    .. math::
         R_{f} = P \, r_f

    Args:
        resp_f: The foliar respiration rate
        whole_crown_gpp: The individual whole crown GPP.
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"resp_f": resp_f},
            size_args={"whole_crown_gpp": whole_crown_gpp},
        )

    return _enforce_2D(whole_crown_gpp * resp_f)


def calculate_gpp_topslice(
    gpp_topslice: NDArray[np.float64],
    whole_crown_gpp: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate gpp topslice.

    Calculates a fixed proportion of the total GPP for the crown that is removed before
    further GPP allocation. This is intended as a helper variable for T Model users to
    simulate processes not included in the T Model such as root exudation or active
    nutrient servicing for mycorriza fungi.

    .. NOTE::

        This is a naive calculation method that is not part of the T model. If values
        for GPP topslice are zero it will have no impact on the T Model calculations.

    Args:
        gpp_topslice: The portion of GPP to remove before allocation.
        whole_crown_gpp: The individual whole crown GPP.
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"gpp_topslice": gpp_topslice},
            size_args={"whole_crown_gpp": whole_crown_gpp},
        )

    return _enforce_2D(whole_crown_gpp * gpp_topslice)


def calculate_reproductive_tissue_respiration(
    resp_rt: NDArray[np.float64],
    reproductive_tissue_mass: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate reproductive tissue respiration.

    Calculates the total reproductive tissue respiration (:math:`R_{rt}`) given the
    reproductive tissue mass (:math:`M_rt`) and the reproductive tissue respiration rate
    of the plant functional type (:math:`r_{rt}`).

    NOTE: This function is not part of the original T Model, but is included here to
    allow for the calculation of reproductive tissue respiration in the same way as
    sapwood respiration.

    .. math::
         R_{rt} = M_rt \, r_rt

    Args:
        resp_rt: The reproductive tissue respiration rate
        reproductive_tissue_mass: The stem reproductive tissue mass.
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"resp_rt": resp_rt},
            size_args={"reproductive_tissue_mass": reproductive_tissue_mass},
        )

    return _enforce_2D(reproductive_tissue_mass * resp_rt)


def calculate_fine_root_respiration(
    zeta: NDArray[np.float64],
    sla: NDArray[np.float64],
    resp_r: NDArray[np.float64],
    foliage_mass: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate fine root respiration.

    Calculates the total fine root respiration (:math:`R_{r}`) given the individual
    foliage mass (:math:`W_f`), along with the fine root respiration rate (:math:`r_r`),
    the ratio of fine root mass to foliage area (:math:`\zeta`) and the specific leaf
    area (:math:`\sigma`) :cite:p:`{see Equation 13, }Li:2014bc`

    .. math::
         R_{r} = \zeta \sigma W_f r_r

    Args:
        zeta: The ratio of fine root mass to foliage area of the PFT.
        sla: The specific leaf area of the PFT.
        resp_r: The respiration rate of fine roots of the PFT.
        foliage_mass: The individual foliage mass.
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={
                "zeta": zeta,
                "sla": sla,
                "resp_r": resp_r,
            },
            size_args={"foliage_mass": foliage_mass},
        )

    return _enforce_2D(zeta * sla * foliage_mass * resp_r)


def calculate_net_primary_productivity(
    yld: NDArray[np.float64],
    whole_crown_gpp: NDArray[np.float64],
    foliar_respiration: NDArray[np.float64],
    fine_root_respiration: NDArray[np.float64],
    sapwood_respiration: NDArray[np.float64],
    reproductive_tissue_respiration: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate net primary productivity.

    The net primary productivity (NPP, :math:`P_{net}`) is calculated as a plant
    functional type specific yield proportion (:math:`y`) of the total GPP (:math:`P`)
    for the individual minus respiration (:math:`R_m`), as the sum of the respiration
    costs for foliage  (:math:`R_f`), fine roots  (:math:`R_r`), sapwood
    (:math:`R_s`), and reproductive tissue (:math:`R_{rt}`).

    .. math::
        P_{net} = y (P - R_m) = y (P - W_{\cdot s} r_s - \zeta \sigma W_f r_r - W_f r_f
        - P r_{rt})

    Note that this differs from Equation 13 of :cite:t:`Li:2014bc`, which does not
    include a term for foliar respiration or reproductive tissue respiration.
    :cite:t:`Li:2014bc` remove foliar respiration as a fixed proportion of potential GPP
    as the first step in their calculations. The approach here is equivalent but allows
    the foliar respiration to vary between plant functional types. :cite:t:`Li:2014bc`
    do not include reproductive tissue respiration in their calculations.

    Args:
        yld: The yield proportion.
        whole_crown_gpp: The total GPP for the crown.
        foliar_respiration: The total foliar respiration.
        fine_root_respiration: The total fine root respiration
        sapwood_respiration: The total sapwood respiration.
        reproductive_tissue_respiration: The total reproductive tissue respiration.
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"yld": yld},
            size_args={
                "whole_crown_gpp": whole_crown_gpp,
                "foliar_respiration": foliar_respiration,
                "fine_root_respiration": fine_root_respiration,
                "sapwood_respiration": sapwood_respiration,
                "reproductive_tissue_respiration": reproductive_tissue_respiration,
            },
        )

    return _enforce_2D(
        yld
        * (
            whole_crown_gpp
            - foliar_respiration
            - fine_root_respiration
            - sapwood_respiration
            - reproductive_tissue_respiration
        )
    )


def calculate_foliage_turnover(
    tau_f: NDArray[np.float64],
    foliage_mass: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate turnover costs for foliage.

    This function calculates the costs associated with the turnover of foliage. This is
    calculated from the total foliage mass of individuals (:math:`W_f`), and the
    turnover times of foliage (:math:`\tau_f`) of the plant functional type
    :cite:p:`{see Equation 15, }Li:2014bc`.

    .. math::

        T = W_f \left( \frac{1}{\tau_f} \right)

    Args:
        tau_f: The turnover time of foliage
        foliage_mass: The foliage mass
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"tau_f": tau_f},
            size_args={"foliage_mass": foliage_mass},
        )

    return _enforce_2D(foliage_mass * (1 / tau_f))


def calculate_fine_root_turnover(
    sla: NDArray[np.float64],
    zeta: NDArray[np.float64],
    tau_r: NDArray[np.float64],
    foliage_mass: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate turnover costs.

    This function calculates the costs associated with the turnover of fine roots. This
    is calculated from the total foliage mass of individuals (:math:`W_f`), along with
    the specific leaf area (:math:`\sigma`) and fine root mass to foliar area ratio
    (:math:`\zeta`) and the turnover time of fine roots (:math:`\tau_r`) of the plant
    functional type :cite:p:`{see Equation 15, }Li:2014bc`.

    .. math::

        T = W_f \left(\frac{ \sigma \zeta}{\tau_f} \right)

    Args:
        sla: The specific leaf area
        zeta: The ratio of fine root mass to foliage area.
        tau_r: The turnover time of fine roots
        foliage_mass: The foliage mass
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"sla": sla, "zeta": zeta, "tau_r": tau_r},
            size_args={"foliage_mass": foliage_mass},
        )

    return _enforce_2D(foliage_mass * (sla * zeta / tau_r))


def calculate_reproductive_tissue_turnover(
    m_rt: NDArray[np.float64],
    tau_rt: NDArray[np.float64],
    validate: bool = True,
) -> NDArray[np.float64]:
    r"""Calculate reproductive tissue turnover costs.

    This function calculates the costs associated with the turnover of reproductive
    tissue. This is calculated from the total reproductive tissue mass
    (:math:`m_rt`), along with the turnover time of reproductive tissue
    (:math:`\tau_rt`).

    .. math::

        T_rt = m_rt \left( \frac{1}{\tau_rt}\right)

    Args:
        m_rt: The mass of reproductive tissue
        tau_rt: The turnover time of reproductive tissue
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={"tau_rt": tau_rt},
            size_args={"m_rt": m_rt},
        )

    return _enforce_2D(m_rt * (1 / tau_rt))


def calculate_reproductive_tissue_mass(
    foliage_mass: NDArray[np.float64],
    p_foliage_for_reproductive_tissue: NDArray[np.float64],
) -> NDArray[np.float64]:
    r"""Calculate reproductive tissue mass.

    This function calculates the mass of reproductive tissue (:math:`m_rt`) as a fixed
    proportion of the total foliage mass (:math:`W_f`) of individuals.

    .. math::

        m_rt = p_{f_rt} W_f

    Args:
        foliage_mass: The foliage mass
        p_foliage_for_reproductive_tissue: The proportion of foliage mass that is
            reproductive tissue
        validate: Boolean flag to suppress argument validation
    """

    return _enforce_2D(p_foliage_for_reproductive_tissue * foliage_mass)


def calculate_growth_increments(
    rho_s: NDArray[np.float64],
    a_hd: NDArray[np.float64],
    h_max: NDArray[np.float64],
    lai: NDArray[np.float64],
    ca_ratio: NDArray[np.float64],
    sla: NDArray[np.float64],
    zeta: NDArray[np.float64],
    npp: NDArray[np.float64],
    turnover: NDArray[np.float64],
    reproductive_tissue_turnover: NDArray[np.float64],
    p_foliage_for_reproductive_tissue: NDArray[np.float64],
    dbh: NDArray[np.float64],
    stem_height: NDArray[np.float64],
    validate: bool = True,
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    r"""Calculate growth increments.

    Given an estimate of net primary productivity (:math:`P_{net}`), less associated  
    turnover costs (:math:`T`), the remaining productivity can be allocated to growth
    and hence estimate resulting increments :cite:`Li:2014bc` in:
    
    * the stem diameter (:math:`\Delta D`),
    * the stem mass (:math:`\Delta W_s`), and 
    * the foliar mass (:math:`\Delta W_f`). 
        
    The stem diameter increment can be calculated using the available productivity for
    growth and the rates of change in stem (:math:`\textrm{d}W_s / \textrm{d}t`) and
    foliar masses (:math:`\textrm{d}W_f / \textrm{d}t`): 

    .. math::

        \Delta D = \frac{P_{net} - T}{ \textrm{d}W_s / \textrm{d}t  +
             \textrm{d}W_f / \textrm{d}t}

    The rates of change in stem and foliar mass can be calculated as:

    .. math::
      :nowrap:

      \[
        \begin{align*}
            \textrm{d}W_s / \textrm{d}t &= \frac{\pi}{8} \rho_s D
                \left(a D \left(1 - \frac{H}{H_{m}} + 2 H \right) \right) \\

            \textrm{d}W_f / \textrm{d}t &= L \frac{\pi c}{4 a} \left(a D \left( 1 -
                \frac{H}{H_{m}} + H \right) \right) \frac{1}{\sigma + \zeta}
        \end{align*}
      \]

    given the current stem diameter (:math:`D`) and height (:math:`H`) and the following
    plant functional type traits:

    * the specific leaf area (:math:`\sigma`),
    * the leaf area index (:math:`L`),
    * the wood  density of the PFT (:math:`\rho_s`),
    * the maximum height (:math:`H_{m}`),
    * the initial slope of the height/diameter relationship (:math:`a`),
    * the crown area ratio (:math:`c`), and
    * the ratio of fine root mass to leaf area (:math:`\zeta`).

    The resulting incremental changes in stem mass and foliar mass can then be
    calculated as:

    .. math::
      :nowrap:

      \[
        \begin{align*}
        \Delta W_s &=  \textrm{d}W_s / \textrm{d}t \, \Delta D\\
        \Delta W_f &=  \textrm{d}W_f / \textrm{d}t \, \Delta D
        \end{align*}
      \]

    NOTE: Reproductive tissue is included as an optional additional cost of turnover.
    If the default values are set to zero, it will not impact T Model calculations.

    Args:
        rho_s: Wood density of the PFT
        a_hd: Initial slope of the height/diameter relationship of the PFT
        h_max: Maximum height of the PFT
        lai: Leaf area index of the PFT
        ca_ratio: Crown area ratio of the PFT
        sla: Specific leaf area of the PFT
        zeta: The ratio of fine root mass to foliage area of the PFT
        npp: Net primary productivity of individuals
        turnover: Fine root and foliage turnover cost of individuals
        p_foliage_for_reproductive_tissue: Proportion of foliage mass that is
            reproductive tissue.
        reproductive_tissue_turnover: Reproductive tissue turnover cost of individuals
        dbh: Diameter at breast height of individuals
        stem_height: Stem height of individuals
        validate: Boolean flag to suppress argument validation
    """
    if validate:
        _validate_demography_array_arguments(
            trait_args={
                "rho_s": rho_s,
                "a_hd": a_hd,
                "h_max": h_max,
                "lai": lai,
                "ca_ratio": ca_ratio,
                "sla": sla,
                "zeta": zeta,
            },
            size_args={
                "npp": npp,
                "turnover": turnover,
                "reproductive_tissue_turnover": reproductive_tissue_turnover,
                "p_foliage_for_reproductive_tissue": p_foliage_for_reproductive_tissue,
                "dbh": dbh,
                "stem_height": stem_height,
            },
        )
    # Rates of change in stem and foliar
    dWsdt = (
        np.pi
        / 8
        * rho_s
        * dbh
        * (a_hd * dbh * (1 - (stem_height / h_max)) + 2 * stem_height)
    )

    dWfdt = (
        lai
        * ((np.pi * ca_ratio) / (4 * a_hd))
        * (a_hd * dbh * (1 - stem_height / h_max) + stem_height)
        * ((1 + p_foliage_for_reproductive_tissue) / sla + zeta)
    )

    # Increment of diameter at breast height
    delta_d = _enforce_2D(
        (npp - turnover - reproductive_tissue_turnover) / (dWsdt + dWfdt)
    )

    return (delta_d, dWsdt * delta_d, dWfdt * delta_d)


@dataclass
class StemAllometry(PandasExporter, CohortMethods):
    """Calculate T Model allometric predictions across a set of stems.

    This method calculates predictions of stem allometries for stem height, crown area,
    crown fraction, stem mass, foliage mass and sapwood mass under the T Model
    :cite:`Li:2014bc`, given diameters at breast height for a set of plant functional
    traits.

    Args:
        stem_traits: An instance of :class:`~pyrealm.demography.flora.Flora` or
            :class:`~pyrealm.demography.flora.StemTraits`, providing plant functional
            trait data for a set of stems.
        at_dbh: An array of diameter at breast height values at which to predict stem
            allometry values.
        validate: Boolean flag to suppress argument validation
    """

    array_attrs: ClassVar[tuple[str, ...]] = (
        "dbh",
        "stem_height",
        "crown_area",
        "crown_fraction",
        "stem_mass",
        "foliage_mass",
        "reproductive_tissue_mass",
        "sapwood_mass",
        "crown_r0",
        "crown_z_max",
    )
    count_attr: ClassVar[str] = "_n_stems"

    # Init vars
    stem_traits: InitVar[Flora | StemTraits]
    """ An instance of :class:`~pyrealm.demography.flora.Flora` or 
    :class:`~pyrealm.demography.flora.StemTraits`, providing plant functional trait data
    for a set of stems."""
    at_dbh: InitVar[NDArray[np.float64]]
    """An array of diameter at breast height values at which to predict stem allometry 
    values."""
    validate: InitVar[bool] = True
    """Boolean flag to suppress argument validation."""

    # Post init allometry attributes
    dbh: NDArray[np.float64] = field(init=False)
    """Diameter at breast height (m)"""
    stem_height: NDArray[np.float64] = field(init=False)
    """Stem height (m)"""
    crown_area: NDArray[np.float64] = field(init=False)
    """Crown area (m2)"""
    crown_fraction: NDArray[np.float64] = field(init=False)
    """Vertical fraction of the stem covered by the crown (-)"""
    stem_mass: NDArray[np.float64] = field(init=False)
    """Stem mass (kg)"""
    foliage_mass: NDArray[np.float64] = field(init=False)
    """Foliage mass (kg)"""
    reproductive_tissue_mass: NDArray[np.float64] = field(init=False)
    """Reproductive tissue mass (kg)"""
    sapwood_mass: NDArray[np.float64] = field(init=False)
    """Sapwood mass (kg)"""
    crown_r0: NDArray[np.float64] = field(init=False)
    """Crown radius scaling factor (-)"""
    crown_z_max: NDArray[np.float64] = field(init=False)
    """Height of maximum crown radius (m)"""

    # Information attributes
    _n_pred: int = field(init=False)
    """The number of predictions per stem."""
    _n_stems: int = field(init=False)
    """The number of stems."""

    __experimental__ = True

    def __post_init__(
        self,
        stem_traits: Flora | StemTraits,
        at_dbh: NDArray[np.float64],
        validate: bool,
    ) -> None:
        """Populate the stem allometry attributes from the traits and size data."""

        warn_experimental("StemAllometry")

        # If validation is required, only need to perform validation once to check that
        # the at_dbh values are congruent with the stem_traits inputs. If they are, then
        # all the other allometry function inputs will be too.
        if validate:
            _validate_demography_array_arguments(
                trait_args={"h_max": stem_traits.h_max}, size_args={"at_dbh": at_dbh}
            )

        self.stem_height = calculate_heights(
            h_max=stem_traits.h_max, a_hd=stem_traits.a_hd, dbh=at_dbh, validate=False
        )

        # Broadcast at_dbh to shape of stem height to get congruent shapes
        self.dbh = np.broadcast_to(at_dbh, self.stem_height.shape)

        self.crown_area = calculate_crown_areas(
            ca_ratio=stem_traits.ca_ratio,
            a_hd=stem_traits.a_hd,
            dbh=self.dbh,
            stem_height=self.stem_height,
            validate=False,
        )

        self.crown_fraction = calculate_crown_fractions(
            a_hd=stem_traits.a_hd,
            dbh=self.dbh,
            stem_height=self.stem_height,
            validate=False,
        )

        self.stem_mass = calculate_stem_masses(
            rho_s=stem_traits.rho_s,
            dbh=self.dbh,
            stem_height=self.stem_height,
            validate=False,
        )

        self.foliage_mass = calculate_foliage_masses(
            sla=stem_traits.sla,
            lai=stem_traits.lai,
            crown_area=self.crown_area,
            validate=False,
        )

        self.reproductive_tissue_mass = calculate_reproductive_tissue_mass(
            self.foliage_mass, stem_traits.p_foliage_for_reproductive_tissue
        )

        self.sapwood_mass = calculate_sapwood_masses(
            rho_s=stem_traits.rho_s,
            ca_ratio=stem_traits.ca_ratio,
            stem_height=self.stem_height,
            crown_area=self.crown_area,
            crown_fraction=self.crown_fraction,
            validate=False,
        )

        self.crown_r0 = calculate_crown_r0(
            q_m=stem_traits.q_m, crown_area=self.crown_area, validate=False
        )

        self.crown_z_max = calculate_crown_z_max(
            z_max_prop=stem_traits.z_max_prop,
            stem_height=self.stem_height,
            validate=False,
        )

        # Set the number of observations per stem as the length of axis 1
        self._n_pred = self.crown_z_max.shape[0]
        self._n_stems = stem_traits._n_stems

    def __repr__(self) -> str:
        return (
            f"StemAllometry: Prediction for {self._n_stems} stems "
            f"at {self._n_pred} DBH values."
        )


@dataclass()
class StemAllocation(PandasExporter):
    """Calculate T Model GPP allocation across a set of stems.

    This method calculates the predicted allocation of potential gross primary
    productivity (GPP) for stems under the T Model :cite:`Li:2014bc`, given a set of
    traits for those stems and the stem allometries given the stem size.

    Args:
        stem_traits: An instance of :class:`~pyrealm.demography.flora.Flora` or
            :class:`~pyrealm.demography.flora.StemTraits`, providing plant functional
            trait data for a set of stems.
        stem_allometry: An instance of
            :class:`~pyrealm.demography.tmodel.StemAllometry`
            providing the stem size data for which to calculate allocation.
        gpp: An array of GPP values available to a stem at which to model allocation (kg
            C).
        validate: Boolean flag to suppress argument validation
    """

    array_attrs: ClassVar[tuple[str, ...]] = (
        "whole_crown_gpp",
        "sapwood_respiration",
        "foliar_respiration",
        "fine_root_respiration",
        "npp",
        "foliage_turnover",
        "fine_root_turnover",
        "delta_dbh",
        "delta_stem_mass",
        "delta_foliage_mass",
    )

    # Init vars
    stem_traits: InitVar[Flora | StemTraits]
    """An instance of :class:`~pyrealm.demography.flora.Flora` or 
    :class:`~pyrealm.demography.flora.StemTraits`, providing plant functional trait data
    for a set of stems."""
    stem_allometry: InitVar[StemAllometry]
    """An instance of :class:`~pyrealm.demography.tmodel.StemAllometry`
    providing the stem size data for which to calculate allocation."""
    whole_crown_gpp: NDArray[np.float64]
    """An array of gross primary productivity values (kg C) across the whole of the
    crown of each stem to be allocated to respiration, turnover and growth."""
    validate: InitVar[bool] = True
    """ Boolean flag to suppress argument validation."""

    # Post init allometry attributes
    topslice_whole_crown_gpp: NDArray[np.float64] = field(init=False)
    """The available stem GPP after any topslicing (g C)"""
    sapwood_respiration: NDArray[np.float64] = field(init=False)
    """Allocation to sapwood respiration (g C)"""
    foliar_respiration: NDArray[np.float64] = field(init=False)
    """Allocation to foliar respiration (g C)"""
    reproductive_tissue_respiration: NDArray[np.float64] = field(init=False)
    """Allocation to reproductive tissue respiration (g C)"""
    fine_root_respiration: NDArray[np.float64] = field(init=False)
    """Allocation to fine root respiration (g C)"""
    gpp_topslice: NDArray[np.float64] = field(init=False)
    """GPP removed before allocation for various biological functions (g C)"""
    npp: NDArray[np.float64] = field(init=False)
    """Net primary productivity (g C)"""
    leaf_turnover: NDArray[np.float64] = field(init=False)
    """Allocation to leaf turnover (g C)"""
    fine_root_turnover: NDArray[np.float64] = field(init=False)
    """Allocation to fine root turnover"""
    reproductive_tissue_turnover: NDArray[np.float64] = field(init=False)
    """Allocation to reproductive tissue turnover (g C)"""
    delta_dbh: NDArray[np.float64] = field(init=False)
    """Predicted increase in stem diameter from growth allocation (m)"""
    delta_stem_mass: NDArray[np.float64] = field(init=False)
    """Predicted increase in stem mass from growth allocation (g C)"""
    delta_foliage_mass: NDArray[np.float64] = field(init=False)
    """Predicted increase in foliar mass from growth allocation (g C)"""

    # Information attributes
    _n_pred: int = field(init=False)
    """The number of predictions per stem."""
    _n_stems: int = field(init=False)
    """The number of stems."""

    __experimental__ = True

    def __post_init__(
        self,
        stem_traits: Flora | StemTraits,
        stem_allometry: StemAllometry,
        validate: bool,
    ) -> None:
        """Populate stem allocation attributes from the traits, allometry and GPP."""

        warn_experimental("StemAllocation")

        if validate:
            _validate_demography_array_arguments(
                trait_args={"h_max": stem_traits.h_max},
                size_args={"dbh": stem_allometry.dbh},
                at_size_args={"whole_crown_gpp": self.whole_crown_gpp},
            )

        # Broadcast potential GPP to match trait and size data outputs
        trait_size_shape = np.broadcast_shapes(
            stem_traits.h_max.shape, stem_allometry.dbh.shape
        )
        self.whole_crown_gpp = np.broadcast_to(self.whole_crown_gpp, trait_size_shape)

        self.gpp_topslice = calculate_gpp_topslice(
            gpp_topslice=stem_traits.gpp_topslice,
            whole_crown_gpp=self.whole_crown_gpp,
            validate=False,
        )

        # Topslice GPP
        self.topslice_whole_crown_gpp = self.whole_crown_gpp - self.gpp_topslice

        self.sapwood_respiration = calculate_sapwood_respiration(
            resp_s=stem_traits.resp_s,
            sapwood_mass=stem_allometry.sapwood_mass,
            validate=False,
        )

        self.foliar_respiration = calculate_foliar_respiration(
            resp_f=stem_traits.resp_f,
            whole_crown_gpp=self.topslice_whole_crown_gpp,
            validate=False,
        )

        self.reproductive_tissue_respiration = (
            calculate_reproductive_tissue_respiration(
                resp_rt=stem_traits.resp_rt,
                reproductive_tissue_mass=stem_allometry.reproductive_tissue_mass,
                validate=False,
            )
        )

        self.fine_root_respiration = calculate_fine_root_respiration(
            zeta=stem_traits.zeta,
            sla=stem_traits.sla,
            resp_r=stem_traits.resp_r,
            foliage_mass=stem_allometry.foliage_mass,
            validate=False,
        )

        self.npp = calculate_net_primary_productivity(
            yld=stem_traits.yld,
            whole_crown_gpp=self.topslice_whole_crown_gpp,
            foliar_respiration=self.foliar_respiration,
            fine_root_respiration=self.fine_root_respiration,
            sapwood_respiration=self.sapwood_respiration,
            reproductive_tissue_respiration=self.reproductive_tissue_respiration,
            validate=False,
        )

        self.foliage_turnover = calculate_foliage_turnover(
            tau_f=stem_traits.tau_f,
            foliage_mass=stem_allometry.foliage_mass,
            validate=False,
        )

        self.fine_root_turnover = calculate_fine_root_turnover(
            sla=stem_traits.sla,
            zeta=stem_traits.zeta,
            tau_r=stem_traits.tau_r,
            foliage_mass=stem_allometry.foliage_mass,
            validate=False,
        )

        self.reproductive_tissue_turnover = calculate_reproductive_tissue_turnover(
            m_rt=stem_allometry.reproductive_tissue_mass,
            tau_rt=stem_traits.tau_rt,
            validate=False,
        )

        (self.delta_dbh, self.delta_stem_mass, self.delta_foliage_mass) = (
            calculate_growth_increments(
                rho_s=stem_traits.rho_s,
                a_hd=stem_traits.a_hd,
                h_max=stem_traits.h_max,
                lai=stem_traits.lai,
                ca_ratio=stem_traits.ca_ratio,
                sla=stem_traits.sla,
                zeta=stem_traits.zeta,
                npp=self.npp,
                turnover=self.foliage_turnover + self.fine_root_turnover,
                reproductive_tissue_turnover=self.reproductive_tissue_turnover,
                p_foliage_for_reproductive_tissue=stem_traits.p_foliage_for_reproductive_tissue,
                dbh=stem_allometry.dbh,
                stem_height=stem_allometry.stem_height,
                validate=False,
            )
        )

        # Set the number of observations per stem (one if dbh is 1D, otherwise size of
        # the first axis)
        if self.whole_crown_gpp.ndim == 1:
            self._n_pred = 1
        else:
            self._n_pred = self.whole_crown_gpp.shape[0]

        self._n_stems = stem_traits._n_stems

    def __repr__(self) -> str:
        return (
            f"StemAllocation: Prediction for {self._n_stems} stems "
            f"at {self._n_pred} observations."
        )



"""Some functions in ``pyrealm`` are only well-behaved with given bounds but those
bounds are often a little imprecise and real world data can contain extreme values. As a
result, the bounds checking is deliberately not that intrusive: it warns when a variable
contains out of value issues but leaves it up to the user to assess whether there is
real problem and to adjust input data if needed.

The ``bounds`` module:

* Defines a {class}`~pyrealm.core.bounds.Bounds` dataclass used to define bounds for a
  particular variable.
* Defines a {class}`~pyrealm.core.bounds.BoundsChecker` class with default bounds for
  core variables that acts as a library for bounds checking.
* The main use case is e.g. ``BoundsChecker().check("tc", np.array([10, 1000])``, which
  will check that the alleged temperature data in °C fall within the configured bounds.

A ``BoundsChecker`` class instance is created with a predefined internal dictionary of
default variables and appropriate bounds. However, users can use the
{meth}`~pyrealm.core.bounds.BoundsChecker.update` method to overide defaults or add new
variables by providing a new ``Bounds`` instance.

The {meth}`~pyrealm.core.bounds.BoundsChecker.check` method can then be used to validate
a set of values against the configured bounds for a given variable name. The ``check``
method returns the input variables, to allow values to be checked while being assigned
to an attribute.
"""  # noqa: D205

from dataclasses import dataclass
from typing import Any, ClassVar
from warnings import warn

import numpy as np
from numpy.typing import NDArray


@dataclass
class Bounds:
    """Bounds checking dataclass for variables."""

    var_name: str
    """A variable name, typically the form used in function arguments."""
    lower: float
    """A lower bound on sensible values."""
    upper: float
    """An upper bound on sensible values."""
    interval_type: str
    """The interval type of the constraint ('[]', '()', '[)', '(]')."""
    unit: str
    """A string giving the expected units."""

    def __post_init__(self) -> None:
        """Bounds data validation."""
        if self.interval_type not in BoundsChecker._interval_types:
            raise ValueError(f"Unknown interval type: {self.interval_type}")

        if self.lower >= self.upper:
            raise ValueError(f"Bounds equal or reversed: {self.lower}, {self.upper}")


class BoundsChecker:
    """A bounds checker for input variables.

    The class provides a library of  {class}`~pyrealm.core.bounds.Bounds` instances for
    core variables, keyed by the
    {attr}`Bounds.var_name<pyrealm.core.bounds.Bounds.var_name>` attribute. The table is
    populated from default values when a ``BoundsChecker`` instance is created but can
    be updated and extended by assigning new ``Bounds`` instances to existing or new
    variable name keys using the ``update`` method.
    """

    # TODO - think about these argument names - some unnecessarily terse.
    _defaults: tuple[tuple[str, float, float, str, str], ...] = (
        ("tc", -25, 80, "[]", "°C"),
        ("vpd", 0, 10000, "[]", "Pa"),
        ("co2", 0, 1000, "[]", "ppm"),
        ("patm", 30000, 110000, "[]", "Pa"),
        ("fapar", 0, 1, "[]", "-"),
        ("ppfd", 0, 3000, "[]", "µmol m-2 s-1"),
        ("theta", 0, 0.8, "[]", "m3 m-3"),
        ("rootzonestress", 0, 1, "[]", "-"),
        ("aridity_index", 0, 50, "[]", "-"),
        ("mean_growth_temperature", 0, 50, "[]", "-"),
        ("rh", 0, 1, "[]", "-"),
        ("lat", -90, 90, "[]", "°"),
        ("sf", 0, 1, "[]", "-"),
        ("pn", 0, 1000, "[]", "mm day-1"),
        ("kWm", 0, 1e4, "[]", "mm"),
        ("leaf_area_index", 0, 20, "[]", "-"),
        ("solar_elevation", -90, 90, "[]", "degrees"),
    )
    """Default bounds data for core forcing variables."""

    _interval_types: ClassVar[dict[str, tuple[np.ufunc, np.ufunc]]] = {
        "()": (np.greater, np.less),
        "[]": (np.greater_equal, np.less_equal),
        "(]": (np.greater, np.less_equal),
        "[)": (np.greater_equal, np.less),
    }
    """Dictionary of numpy function pairs for testing interval types."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._data: dict[str, Bounds] = {}

        for var in self._defaults:
            var_bounds = Bounds(*var)
            self._data[var_bounds.var_name] = var_bounds

    def update(self, bounds: Bounds) -> None:
        """Update or add bounds data.

        The {attr}`Bounds.var_name<pyrealm.core.bounds.Bounds.var_name>` attribute of
        the provided ``Bounds`` instance is used to update an existing entry for the
        name or add checking for a new name.

        Args:
            bounds: A Bounds instance.
        """

        self._data[bounds.var_name] = bounds

    def check(self, var_name: str, values: NDArray) -> NDArray:
        r"""Check inputs fall within bounds.

        This method checks whether the provided values fall within the bounds specified
        for the given variable name and issues a warning when this is not the case. If
        the ``BoundsChecker`` class has not been configured the variable name then a
        warning will be given about lack of bounds checking. The method returns the
        input values, so that the method can be used as a pass through validator for
        assigning attributes.

        Args:
            var_name: The variable name
            values: An np.ndarray

        Returns:
            The input values.

        Examples:
            >>> vals = np.array([-15, 20, 30, 124], dtype=float)
            >>> bounds_checker = BoundsChecker()
            >>> bounds_checker.check("temp", vals)
            array([-15.,  20.,  30., 124.])
        """

        var_bounds = self._data.get(var_name)

        if var_bounds is None:
            warn(
                f"Variable '{var_name}' is not configured in the bounds checker. "
                "No bounds checking performed."
            )
            return values

        # Get the interval functions
        lower_func, upper_func = self._interval_types[var_bounds.interval_type]

        # Do the input values contain out of bound values?
        out_of_bounds = np.logical_xor(
            lower_func(values, var_bounds.lower),
            upper_func(values, var_bounds.upper),
        )

        if np.any(out_of_bounds):
            warn(
                f"Variable '{var_name}' ({var_bounds.unit}) contains values outside "
                f"the expected range ({var_bounds.lower},{var_bounds.upper}). "
                "Check units?"
            )

        return values
